/* this file is a part of amp software, (C) tomislav uzelac 1996,1997

	Origional code by: tomislav uzelac
	Modified by:
	* Andrew Richards - moved code from audio.c

 */

#include "audio.h"

#include <sys/types.h>

#if defined(OS_NetBSD)
  #include <sys/ioctl.h>
#else
  #include <sys/stropts.h>
#endif

#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/audioio.h>
#include <multimedia/libaudio.h>       /* Nasty hack for /usr/demo/SOUND */


/* declare these static to effectively isolate the audio device */

static audio_fd;
static audioctl_fd;
static audio_info_t auinfo;


Audio_hdr   audioinfo;
Audio_info  audiostate;

/* needs to set the play format of the data...  don't know how on this
   architecture -- Chutt */

void audioSetFormat(int signedness, int big, int sixteen)
{
}

/* audioOpen() */
/* should open the audio device and perform any special initialization */
/* returns the file descriptior of the audio device										 */

int
audioOpen(int frequency, int stereo, int volume)
{
	int res;
        int port;

	if ((audio_fd = open("/dev/audio",O_RDWR))==-1) {
		fprintf(stderr," unable to open the audio device\n");
		return 1;
	}
	if ((audioctl_fd = open("/dev/audioctl",O_RDWR))==-1) {
	        close(audio_fd);
		fprintf(stderr," unable to open the audioctl device\n");
		return 1;
	}

	audio_pause(audio_fd);
	audio_flush(audio_fd);

	/*DB(audio, fprintf(stderr,"Audio device opened on %d\n",audio_fd) ); */

	if ((audio_get_play_config(audioctl_fd, &audioinfo) !=
	     AUDIO_SUCCESS) ) {
	   fprintf(stderr,"Cound not get encoding information\n");
	   return 1;
	}
#if 0
	if (ioctl(audio_fd,AUDIO_GETINFO,&auinfo)<0) {
		fprintf(stderr,"Unable to get audio info\n");
		return 1;
	}
#else
	AUDIO_INITINFO(&auinfo);
#endif
	audiostate.play.port = 1;
	audiostate.play.eof = 0;
	audiostate.play.error = 0;
	audiostate.play.samples = 0;
	(void) audio_setinfo(audioctl_fd, &audiostate);

	port = AUDIO_HEADPHONE;
	audio_set_play_port(audioctl_fd, &port);

#if 0
	auinfo.play.precision=16;
	auinfo.play.encoding=AUDIO_ENCODING_LINEAR;
	auinfo.play.channels=(stereo ? 2 : 1); 
	DB(audio, fprintf(stderr,"setting sample rate to %d Hz",frequency) );
	auinfo.play.sample_rate=frequency;
	if (ioctl(audio_fd,AUDIO_SETINFO,&auinfo)<0) {
		fprintf(stderr,"Unable to set audio info\n");
		return 1;
	}
#endif

	if (audio_get_play_config(audio_fd, &audioinfo)) {
	        fprintf(stderr," unable to get audio info\n");
		return 1;
	}

	audioinfo.sample_rate = frequency;
	audioinfo.channels = (stereo ? 2 : 1);
	audioinfo.samples_per_unit = 1;
	audioinfo.bytes_per_unit = 2;
	audioinfo.encoding = AUDIO_ENCODING_LINEAR;
	audioinfo.data_size = AUDIO_UNKNOWN_SIZE;

	/* Initialize audio device */
	if (audio_set_play_config(audio_fd, &audioinfo)) {
	   fprintf(stderr," unable to set audio info\n");
	   return 1;
	}

	audio_resume(audio_fd);

	return 0;
}


/* audioSetVolume */
/* sets the output volume 0-100 */

void
audioSetVolume(int volume)
{
	int res;

	/*DB(audio, fprintf(stderr,"Setting volume to: %d\n",volume) ); */
	AUDIO_INITINFO(&audiostate);
	audiostate.play.gain = (volume*255) / 100;
	(void) audio_set_play_gain(audio_fd, &audiostate);

	if (ioctl(audio_fd,AUDIO_SETINFO,&auinfo)<0)
		fprintf(stderr,"Unable to set sound volume\n");
}


/* audioFlush() */
/* should flush the audio device */

inline void
audioFlush()
{
  /*	DB(audio, fprintf(stderr,"audio: flush %d\n",audio_fd) ); */
}


/* audioClose() */
/* should close the audio device and perform any special shutdown			 */

void
audioClose()
{
	close(audio_fd);
}


/* audioWrite */
/* writes count bytes from buffer to the audio device */
/* returns the number of bytes actually written */

int inline
audioWrite(char *buffer, int count)
{
	return(write(audio_fd,buffer,count));
}

/* Let buffer.c have the audio descriptor so it can select on it. This means	*/
/* that the program is dependent on an file descriptor to work. Should really */
/* move the select's etc (with inlines of course) in here so that this is the */
/* ONLY file which has hardware dependent audio stuff in it										*/

int
getAudioFd()
{
	return(audio_fd);
}
