/* this file is a part of amp software, (C) tomislav uzelac 1996,1997

	Origional code by: tomislav uzelac
	Modified by:
	* Dan Nelson - BSD mods.
	* Andrew Richards - moved code from audio.c and added mixer support etc
        * Isaac Richards - emp3 stuff 
 */

/* Support for Linux and BSD sound devices */

#include "audio.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>


#ifdef HAVE_MACHINE_SOUNDCARD_H
#include <machine/soundcard.h>
#else
#include <linux/soundcard.h>
#endif

#ifdef HAVE_SOUND
#include <esd.h>
#endif

/* optimal fragment size */

int AUSIZ = 32768; 

/* declare these static to effectively isolate the audio device */

static int audio_fd;
static int mixer_fd;
static int volumeIoctl;

static int use_sounD = FALSE;

static int play_format = AFMT_S16_LE;



void audioSetFormat(int signedness, int big, int sixteen)
{
   if (!sixteen) {
      if (signedness) {
      	 play_format = AFMT_S8;
      }
      else {
         play_format = AFMT_U8;
      }
   }
   else {
      if (signedness) {
         if (big) {
            play_format = AFMT_S16_BE;
         }
         else {
            play_format = AFMT_S16_LE;
         }
      }
      else {
         if (big) {
            play_format = AFMT_U16_BE;
         }
         else {
            play_format = AFMT_U16_LE;
         }
      }
   }
}

/* audioOpen() */
/* should open the audio device, perform any special initialization		 */
/* Set the frequency, no of channels and volume. Volume is only set if */
/* it is not -1 */

int
audioOpen(int frequency, int stereo, int volume)
{
	int supportedMixers;

#ifdef HAVE_SOUND
        esd_format_t flags = ESD_STREAM | ESD_PLAY;

        if (!stereo)
	   flags |= ESD_MONO;
	else
	   flags |= ESD_STEREO;
	if ((play_format == AFMT_S8) || (play_format == AFMT_U8))
	   flags |= ESD_BITS8;
	else
	   flags |= ESD_BITS16;
	use_sounD = esd_play_stream(flags, frequency);
        if (use_sounD <= 0) 
           use_sounD = 0; 
	else {
           AUSIZ = ESD_BUF_SIZE; 
           return 0;
        }
#endif
	if ((audio_fd = open("/dev/dsp", O_WRONLY, 0)) == -1) {
		fprintf(stderr,"Unable to open the audio device\n");
		return 1;
	}

	if (ioctl(audio_fd, SNDCTL_DSP_SETFMT,&play_format) < 0) {
		fprintf(stderr,"Unable to set required audio format\n");
		return 1;
	}
	if ((mixer_fd=open("/dev/mixer",O_RDWR)) == -1)
		fprintf(stderr,"Unable to open mixer device\n");

	if (ioctl(mixer_fd, SOUND_MIXER_READ_DEVMASK, &supportedMixers) == -1) {
		fprintf(stderr,"Unable to get mixer info assuming master volume\n");
		volumeIoctl=SOUND_MIXER_WRITE_VOLUME;
	} else {
		if ((supportedMixers & SOUND_MASK_PCM) != 0)
			volumeIoctl=SOUND_MIXER_WRITE_PCM;
		else
			volumeIoctl=0;
	}

	/* Set 1 or 2 channels */
	if (ioctl(audio_fd, SNDCTL_DSP_STEREO, &stereo) < 0) {
		fprintf(stderr,"Unable to set stereo/mono\n");
		return 1;
	}
	/* Set the output frequency */
	if (ioctl(audio_fd, SNDCTL_DSP_SPEED, &frequency) < 0) {
		fprintf(stderr,"Unable to set frequency: %d\n",frequency);
		return 1;
	}

	if (volume != -1)
		audioSetVolume(volume);

/*	if (ioctl(audio_fd, SNDCTL_DSP_SETFRAGMENT, &AUSIZ) == -1) {
		fprintf(stderr,"Unable to get fragment size\n");
		return 1;
	}  */
	return 0;
}


/* audioSetVolume - only code this if your system can change the volume while */
/*									playing. sets the output volume 0-100 */

void
audioSetVolume(int volume)
{
/*	DB(audio, fprintf(stderr,"Setting volume to: %d\n",volume); ) */

   if (use_sounD)
      return; 
   volume=(volume<<8)+volume;
      if ((mixer_fd != -1) && (volumeIoctl!=0))
         if (ioctl(mixer_fd, volumeIoctl, &volume) < 0)
            fprintf(stderr,"Unable to set sound volume\n");
}


/* audioFlush() */
/* should flush the audio device */

inline void
audioFlush()
{
/*	DB(audio, fprintf(stderr,"audio: flush %d\n",audio_fd) ); */
   if (use_sounD)
      return;
   if (!audio_fd)
      return;
   if (ioctl(audio_fd, SNDCTL_DSP_RESET, 0) == -1)
	fprintf(stderr,"Unable to reset audio device\n");
}


/* audioClose() */
/* should close the audio device and perform any special shutdown */

void
audioClose()
{
#ifdef HAVE_SOUND
   if (use_sounD) {
      close(use_sounD);
      use_sounD = 0;
      return;
   }
#endif 
	close(audio_fd);
	if (mixer_fd != -1)
		close(mixer_fd);
/*	DB(audio, fprintf(stderr,"audio: closed %d\n",audio_fd) ); */
   audio_fd = 0;
   play_format = AFMT_S16_LE;
}


/* audioWrite */
/* writes count bytes from buffer to the audio device */
/* returns the number of bytes actually written */

int
audioWrite(char *buffer, int count)
{
/*	DB(audio, fprintf(stderr,"audio: Writing %d bytes to audio descriptor %d\n",count,getAudioFd()) ); */
#ifdef HAVE_SOUND
   if (use_sounD)
      return write(use_sounD, buffer, count);
#endif
	return(write(audio_fd,buffer,count));
}


/* Let buffer.c have the audio descriptor so it can select on it. This means	*/
/* that the program is dependent on a file descriptor to work. Should really */
/* move the select's etc (with inlines of course) in here so that this is the */
/* ONLY file which has hardware dependent audio stuff in it										*/

int
getAudioFd()
{
   if (use_sounD)
      return use_sounD;
   return(audio_fd);
}
