#include "emusic.h"

int playing = FALSE;
int paused = FALSE;
int directAccess = FALSE;
int loop_mode = TRUE;
int repeat_mode = FALSE;
int playtype = FORWARD;
int cur_second = 0;
int seconds_timer = 0;
int num_seconds = 0;
PlayList *playlist, *cdlist, *songlist, *shufflelist;
struct playlist_item *cur_item;
int end_of_song = FALSE;
int dumbcdmode = FALSE;

int end_song = FALSE, stop_song = FALSE;
int shuffled = FALSE;

pthread_mutex_t player_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_t player_thread;

void seek_to_second(int second) 
{
   pthread_mutex_lock(&player_lock);
   cur_item->player->seek_to_frame(second); 
   pthread_mutex_unlock(&player_lock);
   seconds_timer = second;
}

int play(int start_second, struct playlist_item *song)
{
   int i = 0;
   if (song) {
     /* print_tag(song);  */
      i = song->player->play(start_second, song->path);
      if (i) {
         fprintf(stderr,"Now Playing with %s: %s\n", song->player->id, song->name);
         alarm(1);
      }
   } 
   return i;
}                                

void play_frame(void)
{
   if (!playing)
      return;
   if (end_song) 
      handle_end();
   else if (stop_song) {
      control(STOP);
   }
   if (end_song || stop_song)
      end_song = stop_song = FALSE;
   return;
}

void pause_it(void)
{
  if (paused) {
      alarm(0);
      pthread_mutex_lock(&player_lock);
  }
  else {
      alarm(1);
      pthread_mutex_unlock(&player_lock);
  }
  cur_item->player->pause_it();
}

void stop(void)
{
   if (!cur_item) {
	return;
   }
   if (paused)
      control(PAUSE);
   cur_item->player->stop();
   pthread_join(player_thread, NULL);
   pthread_mutex_init(&player_lock, NULL);
   cur_second = 0;
   seconds_timer = 0;
   alarm(0);
}

void switchmodes(void)
{
  if ( getnumtracks_cda() == 0) {
     SetMultiButton(playmodebutton, 0);
     return;
  }
  playing = TRUE;
  stop();
  playing = paused = cur_second = FALSE;
  if (dumbcdmode) {
     if (songlist)
        cur_item = songlist->first;
     playlist = songlist;
     dumbcdmode = FALSE;
  }
  else {
     makelistofnums(cdlist, getnumtracks_cda());
     if (cdlist)
        cur_item = cdlist->first;
     playlist = cdlist;
     dumbcdmode = TRUE;
  }
  SetPlaylistListBox(listbox, playlist);
}

void ShuffleIt(void)
{
  int pos;
  if (cur_item)
     pos = cur_item->pos;
  else
     pos = 0;
  if (!shuffled) {
     copylist(songlist, shufflelist);
     shufflelist = shuffle_list(shufflelist);
     playlist = shufflelist;
     setcurposition(playlist, pos);
     shuffled = TRUE;
  }
  else {
     clearplaylist(playlist);
     playlist = songlist;
     setcurposition(playlist, pos);
     if (playlist && playlist->first)
        cur_item = playlist->first;
     else
        cur_item = NULL;
     shuffled = FALSE;
  }
  SetPlaylistListBox(listbox, playlist);
}

void control(int cmd)
{
   switch(cmd) {
      case PLAY:
         if (!paused && cur_item) {
            if (!play(cur_second, cur_item))
	       handle_end();
            num_seconds = cur_item->length;
            playing = TRUE;
            paused = FALSE;
	    break;
         }
      case PAUSE:
	 if (paused) {
            paused = FALSE;
            playing = TRUE;
	    pause_it();
         }
         else if (playing) {
	    if (!use_wave)
	       audioBufferFlush();
            playing = FALSE;
            paused = TRUE;
            pause_it();
         }                                       
         break;
      case STOP:
         if (playing || paused || (cur_item)) {
            stop();
            cur_second = 0;
            playing = FALSE;
            paused = FALSE;
         }
         break;
      case DIRECTACCESS:
	 if (timer_bar->dragging == 0) {
	       cur_second = ((float)timer_bar->percent * (float)num_seconds);
            if (playing || paused) {
               seek_to_second(cur_second);
            }
         }
         break;
      case UPTRACK:
         if (playing) 
            stop();
         if (paused)
            paused = FALSE;
         cur_second = 0;
         if ((cur_item == playlist->last) && (loop_mode))
            cur_item = playlist->first;
         else if (cur_item == playlist->last)
            playing = FALSE;
         else
            cur_item = cur_item->next;
         if (playing)
            control(PLAY);
         break;
      case DNTRACK:
         if (playing)
            stop();
         if (paused)
           paused = FALSE;
         cur_second = 0;
         if ((cur_item == playlist->first) && (loop_mode))
            cur_item = playlist->last;
         else if (cur_item == playlist->first)
            playing = FALSE;
         else
            cur_item = cur_item->prev;
         if (playing)
            control(PLAY);
         break;
      case FIRST:
         if (playing)
            stop();
         if (paused)
            paused = FALSE;
         cur_second = 0;
         cur_item = playlist->first;
         if (playing)
            control(PLAY);
         break;
      case LAST:
         if (playing)
            stop();
         if (paused)
            paused = FALSE;
         cur_second = 0;
         cur_item = playlist->last;
         if (playing)
            control(PLAY);
         break;
      case CUE:
	 cur_second = seconds_timer + 5;
	 if (cur_second > num_seconds)
	    cur_second = num_seconds;
	 if (playing && !paused)
	    seek_to_second(cur_second);
	 break;
      case REV:
	 cur_second = seconds_timer - 5;
	 if (cur_second < 0)
	    cur_second = 0;
	 if (playing && !paused)
	    seek_to_second(cur_second);
	 break;
   }
}

void getlist(char *list)
{
   char temp[512];
   if (list[0] == '~') {
      strcpy(temp,getenv("HOME"));
      strncat(temp,&list[1],strlen(list)-1);
   }
   else
      strcpy(temp,list);
   loadplaylist(temp);
   if (playtype == SHUFFLE)
      ShuffleIt();
}

void handle_end(void)
{
  end_of_song = TRUE;
  if (!cur_item)
     return;
  if (repeat_mode) {
     control(PLAY);
  }
  else switch (playtype) {
     case FORWARD:
        control(UPTRACK);
        break;
     case REVERSE:
        control(DNTRACK);
        break;
     case SHUFFLE:
        control(UPTRACK);
       break;
     case REPEAT:
       control(PLAY);
       break;
     default:
         control(UPTRACK);
	 break;
  }
  end_of_song = FALSE;
}

void SecondTimer(int sig)
{
   seconds_timer++;
   alarm(1);
}

int get_second(void)
{
   if (!cur_item)
      return 0;
   return seconds_timer;
}

int get_length(struct playlist_item *song)
{
   if (!song)
      return 0;
   return song->player->get_length(song->path);
}

void set_name(struct playlist_item *song)
{
   if (!song)
      return;
   song->player->set_name(song);
}
