#include "emusic.h"

char *temptheme = NULL;
int save_on_exit = FALSE;

void DefaultsEmusic(void)
{
   sprintf(p_list,"%s/.emusic/list",getenv("HOME"));
   songlist = malloc(sizeof(PlayList));
   songlist->first = songlist->last = NULL;
   playlist = songlist;
   cdlist = malloc(sizeof(PlayList));
   cdlist->first = cdlist->last = NULL;
   shufflelist = malloc(sizeof(PlayList));
   shufflelist->first = shufflelist->last = NULL;
   need_to_reload_theme = FALSE;
   themepath = malloc(10);
   strcpy(themepath, "DEFAULT");
   default_path = malloc(80);
   strcpy(default_path, EMUSICDIR);
   strcat(default_path, "themes/winamp-base");
}

void ConfigMain(char *left, char *right)
{
   if (!strncmp("BEGIN",left,5)); 
   else if (!strncmp("END",left,3));
   else if (!strncmp("THEME", left, 5)) {
      if (!themepath) {
         themepath = malloc(strlen(right)+1);
         strcpy(themepath, right);
      }
      if (temptheme) {
         themepath = realloc(themepath, strlen(temptheme)+1);
	 strcpy(themepath, temptheme);
	 free(temptheme);
	 temptheme = NULL;
      }
   }
   else if (!strncmp("PLAYLIST", left, 8)) 
      strcpy(p_list, right);
   else if (!strncmp("LOOPMODE", left, 8))
      loop_mode = parsebool(right);
   else if (!strncmp("DEFAULT_SKIN_PATH", left, 16)) {
      default_path = realloc(default_path, strlen(right) + 1);
      strcpy(default_path, right);
   }
   else if (!strncmp("SAVE_ON_CLOSE", left, 13))
      save_on_exit = parsebool(right);	   
}

static void loop(void)
{
  unsigned long sleeptime;
  XEvent ev;
  Atom wm_del_win = 0;
  if (textmode)
     control(PLAY);
  else 
      wm_del_win = XInternAtom(disp, "WM_DELETE_WINDOW", False);
  for(;;) {
     if (need_to_reload_theme) {
        Unload_Theme(TRUE, newthemepath);
        free(newthemepath);
     }
     play_frame();
     gettimeofday(&tv, NULL);
     lastDraw = thisDraw;
     thisDraw = tv.tv_usec / 10000 + (tv.tv_sec*100);
     if (lastDraw == 0) lastDraw = thisDraw;
     sleeptime = (7 - (thisDraw - lastDraw)) * 10000;
     if ((sleeptime >= 0) && (70000 >= sleeptime)) usleep(sleeptime);
     gettimeofday(&tv, NULL);
     thisDraw = tv.tv_usec / 10000 + (tv.tv_sec*100);
     if (!textmode) {
	if (MainWin->visible)
	   Display_Main();
        while (XPending(disp)) {
           XNextEvent(disp, &ev);
	   switch(ev.type) {
	     case MotionNotify:
		if (MainWin)
		   Event_Main(ev);
		if (PlayWin && PlayWin->visible)
	           Event_PlayEditor(ev);
		if (FileWin && FileWin->visible)
		   EventFileBox(FileWin, ev);
	        if (moving) {
	           MoveWidget(MainWin, NULL, ev.xmotion.x_root - move_x, ev.xmotion.y_root - move_y);
	           if (PlayWin->visible)
		     MoveWidget(PlayWin, MainWin, ev.xmotion.x_root - move_x, ev.xmotion.y_root - move_y);
	        }
	        break;
	     case ButtonRelease: 
	     case ButtonPress:
		if (MainWin)
		    Event_Main(ev);
		if (PlayWin && PlayWin->visible)
	            Event_PlayEditor(ev);
		if (FileWin && FileWin->visible)
	            EventFileBox(FileWin, ev);
	        break;
	     case KeyPress:
		if (MainWin)
	           Event_Main(ev);
		if (PlayWin && PlayWin->visible)
		   Event_PlayEditor(ev);
	        if (FileWin && FileWin->visible)
	           EventFileBox(FileWin, ev);
	        break; 
	     case EnterNotify:
	     case LeaveNotify:
	        if (MainWin)
	           Event_Main(ev);
		if (PlayWin && PlayWin->visible)
	           Event_PlayEditor(ev);
		if (FileWin && FileWin->visible)
		   EventFileBox(FileWin, ev);
	        break;
	     case Expose:
	        if (ev.xexpose.window == MainWin->win) {
	           MainWin->exposed = TRUE;
	           if (MainWin->visible) {
		      name_box->drawn = FALSE;
	              Display_Main();
		   }
                }
	        else if (ev.xexpose.window == PlayWin->win) {
	           PlayWin->exposed = TRUE;
		   if (PlayWin->visible)
		      Display_PlayEditor(); 
                }
	        else if (ev.xexpose.window == FileWin->win) {
	           FileWin->exposed = TRUE;
	           if (FileWin->visible)
		      Display_File(FileWin);
                }
	        break;
	     case ClientMessage:
	        if ((ev.xclient.format == 32) && 
		    (ev.xclient.data.l[0] == wm_del_win)) {
	           if (ev.xclient.window == MainWin->win) 
		      do_action(A_EXIT);
	           else if (ev.xclient.window == PlayWin->win) 
		      do_action(A_PLAYLIST_TOGGLE);
		   else if (ev.xclient.window == FileWin->win) 
		      do_action(A_FILE_CANCEL);
		}
		break;
	   default: break;
	  }
        }
     }
     else
        if (!playing) {
           inputFinished = TRUE;
           while (buffer_running) 
             usleep(100);
           return;
        }
  }
}

void displayDisclaimer(void)
{
   printf("This is eMusic %s, (c) 1997,1998 Isaac Richards\n",VERSION);
   printf("  Comments/suggestions?  Find me on efnet as Chutt.\n");
   printf("\n Read the README for details\n");
   printf("  Looking for modules...\n");
}

void displayUsage(void)
{
   printf("\nusage: emusic [options]\n");
   printf("  --no-graphics, -N (path)\tNo GUI, plays file(s) in path and exits. \"path\"\n");
   printf("\t\t\t\tneeds to exist, and can contain wildcards.\n");
   printf("  --recursive, -R         \tIf used with a wildcard in the path of the above\n");
   printf("\t\t\t\toption this'll recursively search through the\n");
   printf("\t\t\t\tpath adding files to the playlist\n\n");
   printf(" --use-wave, -W           \tOutputs to /dev/raw, for eplus.\n");
   printf(" --theme, -T              \tLoads the theme <theme>\n");
   printf(" --convert, -C <path1> <path2>\tConverts skin at path1 to themeball at path2\n");  
   exit(1);
}

void parse_args(int argc, char *argv[])
{
   int i;
   int bad_options = FALSE;
   textmode = FALSE;
   textmode_recursive = FALSE;
   use_wave_t = FALSE;
   for (i = 1; i < argc; i++) {
      if ((!strcmp("--no-graphics", argv[i]) || !strcmp("-N", argv[i])) && ((argc - i) >= 1)) {
	 textmode = TRUE;
	 textmodepath = malloc(strlen(argv[i+1])+1);
	 strcpy(textmodepath, argv[i+1]);
	 i++;
	 loop_mode = FALSE;
      }
      else if ((!strcmp("--recursive", argv[i]) || !strcmp("-R", argv[i])))
	 textmode_recursive = TRUE;
      else if ((!strcmp("--use-wave", argv[i]) || !strcmp("-W", argv[i])))
         use_wave_t = TRUE;
      else if ((!strcmp("--theme", argv[i]) || !strcmp("-T", argv[i])) && ((argc - i) >= 1)) {
         temptheme = malloc(strlen(argv[i+1])+1);
	 strcpy(temptheme, argv[i+1]);
	 i++;
      }
      else if ((!strcmp("--convert", argv[i]) || !strcmp("-C", argv[i])) && ((argc - i) >= 2)) {
         char *skinpath;
         char *savepath;
         skinpath = GetRightPath(argv[i+1]);
         savepath = GetRightPath(argv[i+2]);
         InitX();
         ConvertSkinToThemeball(skinpath, savepath);
         exit(0);
      }
      else { 
         fprintf(stderr,"bad option %s\n", argv[i]);
	 bad_options = TRUE;
      }
   }
   if (bad_options)
      displayUsage();
}

void Shutdown(void)
{
   char *home;
   playing = TRUE;
   control(STOP);
   Unload_Theme(FALSE, NULL);
   if (save_on_exit) {
      home = malloc(1024);
      strcpy(home, getenv("HOME"));
      strcat(home, "/.emusic/list");
      saveplaylist(playlist, home);
      free(home);
   }
   clearplaylist(playlist);
   free(playlist);
   clearplaylist(cdlist);
   free(cdlist);
   clearplaylist(shufflelist);
   free(shufflelist);
   free(themepath);
   free(default_path);
   ShutdownAnalysis();
   exit(0);
}

void Seg(int sig)
{
   fprintf(stderr,"Hrm..  seems you've found a bug.\n");
   fprintf(stderr,"Fix it and fire off a diff to ijr@po.cwru.edu.\n"); 
   Unload_Theme(FALSE, NULL);
   exit(0);
}

void Pipe(int sig)
{
   fprintf(stderr,"Oops, broken pipe.  This could be caused by some other process using /dev/dsp,\n");
   fprintf(stderr,"or by EsounD dying on eMusic...\n");
   Unload_Theme(FALSE, NULL);
   exit(0);
}

void Interrupt(int sig)
{
   Shutdown();
   exit(0);
}

void SetUpSigs(void)
{
   struct sigaction sigact;
#ifndef WITH_DMALLOC 
   sigact.sa_handler = Seg;
   sigemptyset(&sigact.sa_mask);
   sigact.sa_flags = 0;
   sigaction(SIGSEGV, &sigact, 0);
   sigaction(SIGFPE, &sigact, 0);
   sigaction(SIGILL, &sigact, 0);
   sigact.sa_handler = Pipe;
   sigaction(SIGPIPE, &sigact, 0);
   sigact.sa_handler = Interrupt;
#endif
   sigact.sa_handler = SecondTimer;
   sigemptyset(&sigact.sa_mask);
   sigact.sa_flags = 0;
   sigaction(SIGALRM, &sigact, 0);
}

int main(int argc, char *argv[])
{
   char *dir;
   displayDisclaimer();
   SetUpSigs();
   DefaultsEmusic();
   if (argc > 1)
      parse_args(argc, argv);
   else {
      textmode = FALSE;
      use_wave_t = FALSE;
   }
   loadconfig("/.emusic/MAIN");
   dir = malloc(1024);
   strcpy(dir, EMUSICDIR);
   strcat(dir, "/players");
   InitPlayers(dir);
   free(dir);
   loadconfig("/.emusic/MAIN.players"); 
      parse_args(argc, argv);
   if (!textmode) {
      InitX();
      loadconfig(NULL);
      SpawnAnalysis();
   }
   getlist(p_list);
   if (use_wave_t)
      use_wave = TRUE;
   if (textmode)
      GetTextModeList();
   else 
      XSync(disp, True);
   loop();
   Unload_Theme(FALSE, NULL);
   return 0;
}
