#include "emusic.h"

void setcurposition(PlayList *plist, int pos)
{
   struct playlist_item *cur;
   if (!plist || !plist->first) {
      cur_item = NULL;
      return;
   }
   cur = plist->first;
   while (cur) {
     if (cur->pos == pos) {
       cur_item = cur;
       num_seconds = cur_item->length;
       return;
     }
     cur = cur->next;
   }
}

void copylist(PlayList *source, PlayList *dest)
{
   struct playlist_item *cur, *dcur;
   if (!source || !source->first)
      return;
   if (!dest)
      dest = malloc(sizeof(PlayList));
   else 
      clearplaylist(dest);
   dest->length = source->length;
   cur = source->first;
   dest->first = malloc(sizeof(struct playlist_item));
   dcur = dest->first;
   dcur->prev = NULL;
   while (cur) {
      strcpy(dcur->name, cur->name);
      strcpy(dcur->path, cur->path);
      dcur->length = cur->length;
      dcur->pos = cur->pos;
      dcur->tag = cur->tag;
      dcur->player = cur->player;
      dcur->next = malloc(sizeof(struct playlist_item));
      dcur->next->prev = dcur;
      cur = cur->next;
      dcur = dcur->next;
   }
   dcur = dcur->prev;
   if (dcur->next)
      free(dcur->next);
   dcur->next = NULL;
   dest->last = dcur;
   renumber_list(dest, 1);
}

void printlist(PlayList *plist)
{
   struct playlist_item *cur;
   if (!plist || !plist->first) 
      return;
   cur = plist->first;
   while (cur) {
      fprintf(stderr,"%s\n",cur->name);
      cur = cur->next;
   }   
}

void makelistofnums(PlayList *plist, int numbers)
{
   struct playlist_item *cur;
   int x;
   if (!plist)
      plist = malloc(sizeof(PlayList));
   else
      clearplaylist(plist);
   if (!numbers) {
      plist->first = plist->last = NULL;
      return;
   }
   plist->first = malloc(sizeof(struct playlist_item));
   cur = plist->first;
   cur->prev = NULL;
   for (x = 1; x <= numbers; x++) {
      sprintf(cur->path,"/%d.cda",x);
      set_name(cur);
      cur->player = ReturnPlayerPlugin("cd");
      cur->length = get_length(cur);
      cur->next = malloc(sizeof(struct playlist_item));
      cur->next->prev = cur;
      cur = cur->next;
   }
   cur = cur->prev;
   if (cur->next)
      free(cur->next);
   cur->next = NULL;
   plist->last = cur;
   renumber_list(plist, 0);
}

void clearplaylist(PlayList *plist)
{
   struct playlist_item *cur;
   if (!plist || !plist->first)
      return;
   if (plist->last && plist->first) {
      cur = plist->last->prev;
      if (plist->last != plist->first) {
         while (cur) {
            free(cur->next);
	    cur->next = NULL;
            cur = cur->prev;
	    plist->last = cur;
         }
	 free(plist->first);
      }
      else 
	 free(plist->first);
   }
   plist->last = NULL;
   plist->first = NULL;
}

void renumber_list(PlayList *plist, int prefix)
{
   struct playlist_item *cur;
   char temp[256];
   int i = 1;
   cur = plist->first;
   while (cur) {
      cur->pos = i;
      if (prefix) {
         set_name(cur);
	 strcpy(temp, cur->name);
	 sprintf(cur->name,"(%d) - ", i);
         strcat(cur->name,temp);
      }
      i++;
      cur = cur->next;
   }
}

void sort_list(PlayList *plist, int strip)
{
   struct playlist_item *cur, *next;
   int max;
   int notdone;
   char *name;
   if (!plist->first || !plist->first->next)
      return;
   max = plist->last->pos;
   if (strip) {
     cur = plist->first;
     while (cur) {
        name = strrchr(cur->path,'/') + 1;
	strcpy(cur->name, name);
	cur = cur->next;
     }
   }
   cur = plist->first;
   next = cur->next;
   do {
      notdone = 0;
      while (next) {
	 if (strcmp(next->name,cur->name) < 0) {
	    if (cur == plist->first)
	       plist->first = next;
	    if (next == plist->last)
	       plist->last = cur;
	    cur->next = next->next;
	    if (cur->next)
	       cur->next->prev = cur;
	    next->prev = cur->prev;
	    if (next->prev)
	       next->prev->next = next;
	    cur->prev = next;
	    next->next = cur;
	    notdone = 1;
	 }
	 cur = next;
	 next = cur->next;
	 if (!next) break;
      }
      cur = plist->first;
      next = cur->next;
   }
   while (notdone);
   if (strip)
     renumber_list(plist, 1);
}

void delete_one(int number)
{
   struct playlist_item *cur;
   if (!playlist->first)
      return;
   cur = playlist->first;
   while (cur && (cur->pos != number))
      cur = cur->next;
   if (!cur) 
      return;
   if (cur->pos == number) {
      if (cur->next && cur->prev) {
         cur->prev->next = cur->next;
         cur->next->prev = cur->prev;
	 if (cur_item == cur)
	    cur_item = cur->next;
      }
      else if (cur->prev) {
         cur->prev->next = NULL;
	 playlist->last = cur->prev;
	 if (cur_item == cur)
	    cur_item = cur->prev;
      }
      else if (cur->next) {
	 cur->next->prev = NULL;
	 playlist->first = cur->next;
	 if (cur_item == cur)
	    cur_item = cur->next;
      }
      else {
	 playlist->first = NULL;
	 playlist->last = NULL;
	 cur_item = NULL;
      }
      free(cur);
      renumber_list(playlist, 1);
   }
}

void add_one(int number, struct playlist_item *add)
{
   struct playlist_item *cur;
   if (!playlist->first) return;
   cur = playlist->first;
   while (cur && (cur->pos != number))
      cur = cur->next;
   if (cur) {
      if (cur->prev) {
         cur->prev->next = add;
         add->prev = cur->prev;
      }
      else {
	 playlist->first = add;
         add->prev = NULL;
      }
      cur->prev = add;
      add->next = cur;
   } 
   else {
      add->prev = playlist->last;
      playlist->last->next = add;
      add->next = NULL;
      playlist->last = add;
   }
   renumber_list(playlist, 1);
}

void move_one(int number, int direction)
{
   struct playlist_item *cur, *temp;
   int change_current = FALSE;
   if (!playlist->first) return;
   if (direction && (number == playlist->last->pos)) return;
   if (!direction && (number == 1)) return;
   temp = malloc(sizeof(struct playlist_item));
   cur = playlist->first;
   while (cur && (cur->pos != number))
      cur = cur->next;
   if (!cur) return;
   if (cur_item == cur)
      change_current = TRUE;
   strcpy(temp->name, cur->name);
   strcpy(temp->path, cur->path);
   temp->length = cur->length;
   temp->pos = cur->pos;
   temp->player = cur->player;
   temp->tag = cur->tag;
   delete_one(number);
   if (direction) 
      add_one(number+1,temp);   
   else
      add_one(number-1,temp);
   if (change_current)
      cur_item = temp;
   if (cur_item)
      num_seconds = cur_item->length;
}

PlayList *shuffle_list(PlayList *plist)
{
   time_t currentTime;
   PlayList *shuffled;
   struct playlist_item *cur = NULL, *shuf;
   int i, num;
   int used = 1;
   int max;
   if (!plist->first) 
      return plist; 
   max = plist->last->pos;
   currentTime = time(NULL); 
   srand(currentTime);
   shuffled = malloc(sizeof(PlayList));
   shuffled->last = NULL;
   shuffled->first = malloc(sizeof(struct playlist_item));
   shuf = shuffled->first;
   shuf->prev = NULL;
   for (i = 1; i <= max; i++) {
      while (used) {
         num = (float)rand()/(RAND_MAX+1.0) * max + 1;
         cur = plist->first;
         while ((used) && (cur->next)) {
            if (cur->pos == num)
              used = 0;
            else
              if (cur->next)
                 cur = cur->next;
         }
         if (cur->pos == num)
           used = 0;
      }
      strcpy(shuf->name,cur->name);
      strcpy(shuf->path,cur->path);
      shuf->length = cur->length;
      shuf->pos = i;
      shuf->player = cur->player;
      shuf->tag = cur->tag;
      cur->pos = 0;
      shuf->next = malloc(sizeof(struct playlist_item));
      shuf->next->prev = shuf;
      shuf = shuf->next;
      used = 1;
   }
   shuf = shuf->prev;
   if (shuf->next)
      free(shuf->next);
   shuffled->last = shuf;
   shuf->next = NULL;
   clearplaylist(plist);
   free(plist);
   plist = NULL;
   plist = shuffled;
   return shuffled;
}

int loadplaylist(char *file)
{
   FILE *list;
   char *buffer, *temp;
   int i, type = 0;
   int loaded = 0;
   struct playlist_item *cur;
   clearplaylist(playlist);
   if (!(list = fopen(file,"r"))) {
      return 0;
   }
   buffer=malloc(1024);
   if ((temp = strrchr(file,'/')))
      temp = '\0';
   chdir(file);
   playlist->first = malloc(sizeof(struct playlist_item));
   cur = playlist->first;
   cur->prev = NULL;
   for (i = 0; i < 256; i++)
      buffer[i] = '\0';
   while (fgets(buffer,256,list)) {
     if (buffer[0] == '[')
        type = 1;
     if (type) {
       if (strstr(buffer,"NumberOfEntries"))
          continue;
       else 
          buffer = strchr(buffer,'=') + 1;
     }
     loaded = 1;
     strcpy(cur->path,buffer);
     cur->path[(strlen(cur->path)-1)] = '\0';
     cur->player = getfiletype(cur->path);
     cur->player->get_info(cur);
     set_name(cur);
     cur->length = get_length(cur);
     cur->next = malloc(sizeof(struct playlist_item));
     cur->next->prev = cur;
     cur = cur->next;
   }
   if (loaded) {
      cur = cur->prev;
      if (cur->next)
         free(cur->next);
      cur->next = NULL;
      playlist->last = cur;
   }
   else {
      free(playlist->first);
      playlist->first = NULL;
      playlist->last = NULL;
   }
   cur_item = playlist->first;
   if (cur_item)
      num_seconds = cur_item->length;
   fclose(list);
   renumber_list(playlist, 1);
   return 1;
}                        

struct playlist_item *loadonefile(char *path)
{
   struct playlist_item *cur;
   cur = malloc(sizeof(struct playlist_item));
   strcpy(cur->path, path);
   cur->path[(strlen(cur->path)-1)] = '\0';
   cur->length = get_length(cur);
   cur->next = NULL;
   cur->prev = NULL;
   return cur;
}

void addplaylist(char *path)
{
   FILE *list;
   char *buffer, *temp;
   int i;
   struct playlist_item *cur;
   if (!playlist->first) {
      loadplaylist(path);
      return;
   }
   if (!(list = fopen(path,"r"))) {
      return;
   }
   buffer= malloc(256);
   if ((temp = strrchr(path,'/')))
      temp = '\0';
   chdir(path);
   for (i = 0; i < 256; i++)
      buffer[i] = '\0';
   while (fgets(buffer,256,list)) {
     cur = malloc(sizeof(struct playlist_item));
     strcpy(cur->path, buffer);
     cur->path[(strlen(cur->path)-1)] = '\0';
     cur->player = getfiletype(cur->path);
     cur->player->get_info(cur);
     set_name(cur);   
     cur->length = get_length(cur);
     cur->next = NULL;
     cur->prev = NULL;
     if (listbox) {
        add_one(listbox->current_selection, cur);
        listbox->current_selection++;
     }
   }
   fclose(list);
   renumber_list(playlist, 1);
   free(buffer);
   return;
}

int saveplaylist(PlayList *plist, char *file)
{
   FILE *list;
   struct playlist_item *cur;
   if (!(list = fopen(file,"w"))) {
      fprintf(stderr,"Couldn't write playlist\n");
      return 0;
   }
   if (plist->first) {
      cur = plist->first;
      while (cur) {
         fprintf(list,"%s\n",cur->path);
         cur = cur->next;
      }
   }
   fclose(list);
   return 1;
}                    
