#include "emusic.h"

Widget *PlayWin;
ListBox *listbox;
int actiontype;


void Create_PlayWin(void)
{
  if (!PlayWin->state)
     HideWidget(PlayWin);
}

void PlayWin_Callback(void)
{
   char *temp;
   char returnpath[256];
   if (FileWin->okdone) {
      strcpy(returnpath, ReturnTextTextBox(FileWin->returnpath));
      switch (actiontype) {
      case ADD: if (FileWin->cdmode) 
                   cdtolist(returnpath, listbox->current_selection);
                else if (!chdir(returnpath))
                   dirtolist(returnpath,"*", FileWin->recurse);
                else {
                   temp = splitdir(returnpath);
                   if (strlen(returnpath) == 0)
                      dirtolist("/",temp,FileWin->recurse);
                   else
                      dirtolist(returnpath, temp, FileWin->recurse);
                }
                break;
      case SAVE: saveplaylist(playlist, returnpath); break;
      case LOAD: loadplaylist(returnpath); break;
      default: break;
      }
      FileWin->done = TRUE;
      actiontype = 0;
      DisplayListBox(listbox);
   }
}


void Display_PlayEditor(void)
{
  if (PlayWin->exposed) {
     DisplayListBox(listbox);
     PlayWin->exposed = FALSE;
  }
  if (FileWin->visible) 
     return;
}

void CheckToPlay(void)
{
  DisplayListBox(listbox);
  if (!playlist->first) {
      cur_item = NULL;
      playing = FALSE;
  }    
  if (!cur_item || ((listbox->current_selection == cur_item->pos) && playing)) {
      stop();
      if (playing && cur_item)
         control(PLAY);
  }
}

void CreatePlayWin(void)
{
   struct _widget *cur;
   if (!PlayWin)
      return;
   if (!PlayWin->first)
      return;
   deferred = TRUE;
   cur = PlayWin->first;
   while (cur) {
      switch (cur->type) {
      case BACKGROUND: CreateBack(cur->widget);
	               PlayWin->win = ((Background *)cur->widget)->win; 
		       MoveWidget(PlayWin, MainWin, MainWin->x, MainWin->y); 
		       break;
      case BUTTON: CreateButton(cur->widget, PlayWin->win); break;
      case MULTIBUTTON: CreateMultiButton(cur->widget, PlayWin->win); break;
      case SLIDER: CreateSlider(cur->widget, PlayWin->win); break;
      case TEXTBOX: CreateTextBox(cur->widget, PlayWin->win); break;
      case LISTBOX: CreateListBox(cur->widget, PlayWin->win); 
                    listbox = cur->widget; break;
      default: break;
      }
      cur = cur->next;
   }
   PlayWin->deferred = FALSE;
   deferred = FALSE;
   PlayWin->visible = TRUE;
}

void TogglePlayWin(void)
{
   if (PlayWin->visible)
      HideWidget(PlayWin);
   else {
      if (PlayWin->deferred)
	 CreatePlayWin();
      MoveWidget(PlayWin, MainWin, MainWin->x, MainWin->y);
      UnhideWidget(PlayWin);
      if (PlayWin->raised != MainWin->raised)
	 RaiseLowerWidget(PlayWin);
      listbox->current_selection = 1;
      listbox->start = 1;
      DisplayListBox(listbox);
   }
}

void Event_PlayEditor(XEvent ev)
{
   if (FileWin->visible)
      return;
   if (!FindEventWidget(PlayWin, ev)) {
      if (!PlayWin->shaped)
      	 return;
      if ((ev.type == ButtonPress) && (ev.xbutton.window == PlayWin->win)) {
	 move_x = ev.xbutton.x + PlayWin->offset_x;
	 move_y = ev.xbutton.y + PlayWin->offset_y;
	 moving = TRUE;
      }
      else if ((ev.type == ButtonRelease) && 
	       (ev.xbutton.window == PlayWin->win)) {
	 if (ev.xbutton.button == 3) {
	    RaiseLowerWidget(MainWin);
	    RaiseLowerWidget(PlayWin);
	 }
	 else
	    moving = FALSE;
      }
   }
}

void Select_Current(int current)
{
   struct playlist_item *cur;
   if (!playlist->first) { 
     cur_item = NULL;
     return;
   }
   cur = playlist->first;
   while (cur && (cur->pos != current))
      cur = cur->next;
   if (cur) {
      cur_item = cur;
      if (playing) {
	 control(STOP);
         control(PLAY);
      }
   }
}

void Defaults_PlayWin(void)
{
   PlayWin = malloc(sizeof(Widget));
   PlayWin->visible = TRUE;
   PlayWin->raised = TRUE;
   PlayWin->win = root;
   PlayWin->first = NULL;
   listbox = NULL;
}

void ConfigPlayEditor(FILE *file)
{
   char st[256];
   char s1[128], s2[128];
   struct _widget *cur = NULL;
   int end = FALSE;
   if (textmode)
      return;
   Defaults_PlayWin();
   while (!end) {
      if (!fgets(st, 256, file)) {
	 fprintf(stderr,"Unexpected EOF in config file\n");
	 exit(-1);
      }
      if ((st[0] != '#') && (st[0] != '\n') && (st[0] != '\0')) {
	 splitstring(s1, s2, st);
	 XSync(disp, False);
	 if (!strncmp("END", s1, 3))
	    end = TRUE;
         else if (!strncmp("OFFSET", s1, 10))
            getcoords(s2, &PlayWin->offset_x, &PlayWin->offset_y);
	 else if (!strncmp("VISIBLE", s1, 7)) {
	    PlayWin->state = parsebool(s2);
	    deferred = PlayWin->deferred = !PlayWin->state;
	 }
	 else if (!strncmp("BEGIN", s1, 5)) {
	    upstr(s2);
	    if (!PlayWin->first) {
	       PlayWin->first = malloc(sizeof(struct _widget));
	       cur = PlayWin->first;
	    }
	    else {
	       cur->next = malloc(sizeof(struct _widget));
	       cur = cur->next;
	    }
	    PlayWin->last = cur;
	    cur->next = NULL;
	    if (!strncmp("BACKGROUND", s2, 9)) {
	       cur->type = BACKGROUND;
	       cur->widget = malloc(sizeof(Background));
	       ConfigBack(cur->widget, file);
	       PlayWin->win = ((Background *)cur->widget)->win;
	       PlayWin->shaped = ((Background *)cur->widget)->shaped;
	       if (!deferred) 
	          MoveWidget(PlayWin, MainWin, 0, 0);
	    }
	    else if (!strncmp("BUTTON", s2, 6)) {
	       cur->type = BUTTON;
	       cur->widget = malloc(sizeof(Button));
	       ConfigButton(cur->widget, PlayWin->win, file);
	    }
	    else if (!strncmp("LISTBOX", s2, 7)) {
	       cur->type = LISTBOX;
	       cur->widget = malloc(sizeof(ListBox));
	       ConfigListBox(cur->widget, PlayWin->win, file);
	       SetPlaylistListBox(cur->widget, playlist);
	       listbox = cur->widget;
	    }
	 }
      }
   }
   if (!deferred)
      Create_PlayWin();
   else 
      PlayWin->visible = FALSE;
   XSync(disp, False);
   deferred = FALSE;
}
