#include "emusic.h"

int themeloaded = FALSE;
char *temppath = NULL;

#define UNDEF   0
#define MAIN    1
#define EMP3    2
#define PLIST   3
#define FILEWIN 4

void Theme_Callback(void)
{
   char returnpath[512];
   if (FileWin->okdone) {
      strcpy(returnpath, ReturnTextTextBox(FileWin->returnpath));
      newthemepath = realloc(newthemepath, strlen(returnpath)); 
      strcpy(newthemepath, returnpath);
      need_to_reload_theme = TRUE;
   }
}

void DefaultsTheme(void)
{
   imask.r = 255;
   imask.g = 0;
   imask.b = 255;
   moving = deferred = FALSE;
   XSetForeground(disp, fg, WhitePixel(disp, screen));
}

void loadmain(char *left, char *right)
{
   int n;
   int r, g, b;
   if (!strncmp("BEGIN", left, 5))
      DefaultsTheme();
   if (!strncmp("MASKRGB",left,7)) {
      n = sscanf(right,"%i %i %i", &imask.r, &imask.g, &imask.b);
      if (n != 3) {
         fprintf(stderr,"Bad mask color\n");
         exit(-1);
      }
   }
   else if (!strncmp("FOREGROUNDRGB", left, 13)) {
      n = sscanf(right,"%i %i %i", &r, &g, &b);
      if (n != 3) {
         fprintf(stderr,"Bad foregound color\n");
 	 exit(-1);
      }
      XSetForeground(disp, fg, Imlib_best_color_match(imlib, &r, &g, &b));
      XSetBackground(disp, bg, Imlib_best_color_match(imlib, &r, &g, &b));
   }
}

int readtheme(char *path)
{
   char themefile[512];
   char s1[128], s2[128];
   FILE *in;
   char st[256];
   char savepath[512];
   int context = UNDEF;
   getcwd(savepath, 512); 
   chdir(path);
   sprintf(themefile,"%s/MAIN.theme", path);
   if ((in = fopen(themefile, "r")) == NULL) {
      fprintf(stderr,"Error, error, can't find %s\n", themefile);
      return 0;
   }
   Fnlib_add_dir(fnlib, path);
   while (fgets(st, 256, in)) {
     if ((st[0] != '#') && (st[0] != '\n') && (st[0] != '\0')) {
	 splitstring(s1, s2, st);
	 if (!strncmp("BEGIN", s1,5)) {
	    upstr(s2);
	    if (!strncmp("MAIN", s2, 4)) context = MAIN;
	    else if (!strncmp("EMUSIC", s2, 4)) context = EMP3;
	    else if (!strncmp("PLAYEDITWIN", s2, 5)) context = PLIST;
	    else if (!strncmp("FILE",s2,4)) context = FILEWIN;
	    else context = UNDEF;
	 }
	 if (context == MAIN)
	    loadmain(s1, s2);
         else if (context == EMP3)
            ConfigEmp3(in);
	 else if (context == PLIST)
	    ConfigPlayEditor(in);
	 else if (context == FILEWIN)
	    ConfigFile(in);
	 if (!strncmp("END", s1, 3))
	    context = UNDEF;
     }
   }
   chdir(savepath);
   themeloaded = TRUE;
   return 1;
}

int Load_Theme(void)
{
   int i;
   static int number = 1;
   char s1[1024];
   unsigned char buf[16];
   FILE *f;
   if (!themepath)
      return 0;
   if (themeloaded)
      Unload_Theme(TRUE, NULL);
   i = getpid();
   sprintf(s1, "%s",themepath);
   if (!exists(s1)) {
      sprintf(s1,"%s/.emusic/%s",getenv("HOME"),themepath);
      if (exists(s1)) {
         themepath = realloc(themepath, strlen(s1) + 1);
         strcpy(themepath, s1);
      }
      else {
         sprintf(s1,"%s/themes/%s", EMUSICDIR, themepath);
         themepath = realloc(themepath, strlen(s1) + 1);
         strcpy(themepath, s1);
      }
   }
   if (exists(themepath)) {
     if (isfile(themepath)) {
        temppath = malloc(50);
        sprintf(temppath,"/tmp/emusic_theme_%i%i" ,i, number);
        number++;
	mkdir(temppath, S_IRWXU);
	f = fopen(themepath, "r");
	if (f) {
	   fread(&buf[0], 1, 1, f);
	   fread(&buf[1], 1, 1, f);
	   fclose(f);
	   if ((buf[0] == 31) && (buf[1] == 139))
	      sprintf(s1,"gzip -d -c < %s | (cd %s ; tar -xf -)", themepath, temppath);
	   else if (isSkinZip(themepath))
	      sprintf(s1,"unzip %s -d %s > /dev/null", themepath, temppath);
	   else
	      sprintf(s1,"cd %s ; tar -xf %s", temppath, themepath);
	   system(s1);
	   if (isSkinZip(themepath)) {
	      fprintf(stderr,"Attempting to convert skin %s\n", themepath);
	      if (Convert_Skin(temppath))
	         return 0;
	   }
	}
     }
     else if (!isdir(themepath)) 
        return 0;
   }
   else 
      return 0;
   if (temppath)
      return readtheme(temppath);
   else 
      return readtheme(themepath);
}

void Unload_Theme(int really, char *newtheme)
{
   int sucessful;
   char oldtheme[512];
   if (!themeloaded && textmode)
      return;
   if (temppath) {
     rmrf(temppath);
     Fnlib_del_dir(fnlib, temppath);
   }
   else
     Fnlib_del_dir(fnlib, themepath);
   temppath = NULL;
   if (!really || !newtheme)
     return;
   FreeWidget(MainWin);
   FreeStatusWin(statuswin);
   FreeWidget(PlayWin);
   FreeFileBox(FileWin);
   strcpy(oldtheme, themepath);
   themepath = realloc(themepath, strlen(newtheme) + 1);
   strcpy(themepath, newtheme);
   sucessful = Load_Theme();
   if (!sucessful) {
      fprintf(stderr,"couldn't load %s, reverting to old theme\n", newtheme);
      themepath = realloc(themepath, strlen(oldtheme) + 1);
      strcpy(themepath, oldtheme);
      sucessful = Load_Theme();
      if (!sucessful) {
         fprintf(stderr,"couldn't even load the old one, dying\n");
         Unload_Theme(FALSE, NULL);
         exit(-1);
      }
   }
   need_to_reload_theme = FALSE;
}
