#include "../emusic.h"

void HideBack(Background *back)
{
   if (back->visible) {
      XUnmapWindow(disp, back->win);
      back->visible = FALSE;
   }
}

void UnhideBack(Background *back)
{
   if (!back->visible) {
      XMapWindow(disp, back->win);
      back->visible = TRUE;
   }
}

void CreateBack(Background *back)
{
   XSetWindowAttributes xatrib;
   XTextProperty name;
   Atom wm_del_win = XInternAtom(disp, "WM_DELETE_WINDOW", False);
   if (back->shaped)
      xatrib.override_redirect = TRUE;
   else
      xatrib.override_redirect = FALSE;
   back->win = XCreateWindow(disp, root, back->x, back->y, back->w, back->h, 0, depth, InputOutput, visual, CWOverrideRedirect, &xatrib);
   XSelectInput(disp, back->win, ExposureMask | EnterWindowMask | LeaveWindowMask | ButtonMotionMask | ButtonPressMask | ButtonReleaseMask | KeyPressMask);
   XMapWindow(disp, back->win);
   XStringListToTextProperty(&back->name, 1, &name);
   XSetWMName(disp, back->win, &name);
   XSetWindowBackgroundPixmap(disp, back->win, back->back->pixmap);
   XShapeCombineMask(disp, back->win, ShapeBounding, 0, 0, back->back->mask, ShapeSet);
   XClearWindow(disp, back->win);
   XSetWMProtocols(disp, back->win, &wm_del_win, 1);
   XSync(disp, False);
   back->visible = TRUE;
   back->exposed = FALSE;
   Imlib_free_pixmap(imlib, back->back->pixmap);
   Imlib_free_pixmap(imlib, back->back->mask);
}

void DefaultsBack(Background *back)
{
   back->x = back->y = back->w = back->h = 0;
   back->back = malloc(sizeof(ImageDat));
   memset(back->back, 0, sizeof(ImageDat));
   back->state = 1;
   back->exposed = FALSE;
   back->visible = TRUE;
   back->shaped = TRUE;
   back->name = malloc(32);
   strcpy(back->name,"eMusic ");
   strcat(back->name,VERSION);
}

void ConfigBack(Background *back, FILE *file)
{
   char st[256];
   char s1[128], s2[128];
   int end = FALSE;
   DefaultsBack(back);
   while (!end) {
      if (!fgets(st, 256, file)) {
	 fprintf(stderr,"Unexpected EOF in config file\n");
	 exit(-1);
      }
      if ((st[0] != '#') && (st[0] != '\n') && (st[0] != '\0')) {
	 splitstring(s1, s2, st);
	 if (!strncmp("END", s1, 3))
	    end = TRUE;
	 else if (!strncmp("BACKGROUND", s1, 10)) {
	    LoadImage(s2, back->back);
	    back->w = back->back->w;
	    back->h = back->back->h;
	 }
	 else if (!strncmp("VISIBLE", s1, 7))
	    back->state = parsebool(s2);
	 else if (!strncmp("SHAPED", s1, 6))
	    back->shaped = parsebool(s2);
	 else if (!strncmp("NAME", s1, 4))
	    strncpy(back->name, s2, strlen(s2));
      }
   }
   if (!deferred)
      CreateBack(back);
}

void EventBack(Background *back, XEvent ev)
{
}

void FreeBack(Background *back)
{
   if (!back)
      return;
   free(back->name);
   back->name = NULL;
   FreeImageDat(back->back);
   free(back);
   back = NULL;
}
