#include "../emusic.h"

void HideButton(Button *button)
{
   if (button->visible) {
      XUnmapWindow(disp, button->win);
      button->visible = FALSE;
   }
}

void UnhideButton(Button *button)
{
   if (!button->visible) {
      XMapWindow(disp, button->win);
      button->visible = TRUE;
   }
}

void HideMultiButton(MultiButton *mbutton)
{
   if (mbutton->visible && mbutton->current) {
      HideButton(mbutton->current->button);
      mbutton->visible = FALSE;
   }
}

void UnhideMultiButton(MultiButton *mbutton)
{
   if (!mbutton->visible) {
      UnhideButton(mbutton->current->button);
      mbutton->visible = TRUE;
   }
}

void ClearButton(Button *button)
{
   if (button->win) 
      XClearWindow(disp, button->win);
}

void SetButton(Button *button)
{
   if (!button->current)
      button->current = button->normal;
   XSetWindowBackgroundPixmap(disp, button->win, button->current->pixmap);
   XShapeCombineMask(disp, button->win, ShapeBounding, 0, 0, button->current->mask, ShapeSet);
   ClearButton(button);
   button->exposed = FALSE;
}

void SetButtonBG(Button *button)
{
   button->current = button->normal;
   SetButton(button);
}

void ClickButton(Button *button)
{
   if (!button->clicked) 
      button->current = button->normal;
   else if (button->current == button->clicked)
      return;
   else
      button->current = button->clicked;
   SetButton(button);
}

void OverButton(Button *button)
{
   if (!button->over)
      return;
   if (button->current == button->over)
      return;
   button->current = button->over;
   SetButton(button);
}

void NormalButton(Button *button)
{
   if (!button->normal)
      return;
   if (button->current == button->normal)
      return;
   button->current = button->normal;
   SetButton(button);
}

void SetButtonPart(Button *button, int x)
{
   Pixmap p;
   int partHeight;
   ImageDat *i;

   if (!button)
      return;
   i = button->normal;
   
   if (!i)
      return;
   if ((button->currentPart == x) && (x != 0))
      return;
   if (i->h > i->w) {  
      if (i->parts)
         partHeight = i->h / i->parts;
      else
         partHeight = i->h;
      p = XCreatePixmap(disp, button->win, i->w, partHeight, depth);
      XCopyArea(disp, i->pixmap, p, fg, 0, partHeight*x, i->w, partHeight, 0, 0);
      XShapeCombineMask(disp, button->win, ShapeBounding, 0, -x*partHeight, i->mask, ShapeSet);
   }
   else {
      if (i->parts)
         partHeight = i->w / i->parts;
      else
         partHeight = i->w;
      p = XCreatePixmap(disp, button->win, partHeight, i->h, depth);
      XCopyArea(disp, i->pixmap, p, fg, partHeight*x, 0, partHeight, i->h, 0, 0);
      XShapeCombineMask(disp, button->win, ShapeBounding, -x*partHeight, 0, i->mask, ShapeSet);
   }   
   XSetWindowBackgroundPixmap(disp, button->win, p);
   XClearWindow(disp, button->win);
   button->currentPart = x;
   XFreePixmap(disp, p);
}

void CreateButton(Button *button, Window parent)
{
   int h = button->h;
   int w = button->w;
   if (!button)
      return;
   button->parent = parent;
   if (!button->parent)
      button->parent = root;
   if (button->parts) {
      if (button->h > button->w)
         h = button->h / button->parts;
      else 
         w = button->w / button->parts;
   }
   button->win = CreateWin(button->parent, button->x, button->y, w, h);
   XMapWindow(disp, button->win);
   button->normal->parts = button->parts;
   if (!button->parts)
      SetButtonBG(button);
   else
      SetButtonPart(button, button->currentPart);
   if (!button->state)
      HideButton(button);
   if (!button->clicked && !button->over && !button->parts) {
      Imlib_free_pixmap(imlib, button->normal->pixmap);
      Imlib_free_pixmap(imlib, button->normal->mask);
   }
}

void DefaultsButton(Button *button)
{
   button->x = button->x = button->w = button->h = 0;
   button->normal = button->clicked = button->over = NULL;
   button->current = NULL;
   button->currentPart = 0;
   button->parts = 0;
   button->state = 1;
   button->exposed = 0;
   button->visible = TRUE;
   button->action_left = button->action_middle = button->action_right = 0;
}

void ConfigButton(Button *button, Window parent, FILE *file)
{
   char st[256];
   char s1[128], s2[128];
   int end = FALSE;
   DefaultsButton(button);
   button->parent = parent;
   while (!end) {
      if (!fgets(st, 256, file)) {
	 fprintf(stderr,"Unexpected EOF in config file\n");
	 exit(-1);
      }
      if ((st[0] != '#') && (st[0] != '\n') && (st[0] != '\0')) {
	 splitstring(s1, s2, st);
	 if (!strncmp("#", s1, 1)) ;
	 else if (!strncmp("END", s1, 3))
	    end = TRUE;
	 else if (!strncmp("NORMAL", s1, 6)) {
	    button->normal = malloc(sizeof(ImageDat));
	    LoadImage(s2, button->normal);
	    button->w = button->normal->w;
	    button->h = button->normal->h;
	 }
	 else if (!strncmp("CLICKED", s1, 7)) {
	    button->clicked = malloc(sizeof(ImageDat));
	    LoadImage(s2, button->clicked);
	 }
	 else if (!strncmp("MOUSEOVER", s1, 9)) {
	    button->over = malloc(sizeof(ImageDat));
	    LoadImage(s2, button->over);
	 }
	 else if (!strncmp("LOCATION", s1, 8))
	    getcoords(s2, &button->x, &button->y);
	 else if (!strncmp("PARTS", s1, 5))
	    button->parts = atoi(s2);
	 else if (!strncmp("VISIBLE", s1, 7))
	    button->state = parsebool(s2);
	 else if (!strncmp("ACTION_R", s1, 8))
	    button->action_right = parse_action(s2);
	 else if (!strncmp("ACTION_M", s1, 8))
	    button->action_middle = parse_action(s2);
	 else if (!strncmp("ACTION_L", s1, 8))
	    button->action_left = parse_action(s2);
      }
   }
   if (!deferred)
      CreateButton(button, parent);
}

void EventButton(Button *button, XEvent ev)
{
   static int inside = 0;
   if (ev.type == ButtonPress) {
      ClickButton(button);
      inside = TRUE;
   }
   else if (ev.type == EnterNotify) {
      if (ev.xcrossing.state) 
         ClickButton(button);
      else
         OverButton(button);
      inside = TRUE;
   }
   else if (ev.type == LeaveNotify) {
      NormalButton(button);
      inside = FALSE;
   }
   else if (ev.type == ButtonRelease) {
     if (!inside)
        return;
     inside = FALSE;
     NormalButton(button);
     if (ev.xbutton.button == 3 && !button->action_right)
	    ev.xbutton.button = 1;
     else if (ev.xbutton.button == 2 && !button->action_middle)
        ev.xbutton.button = 1;
     if (ev.xbutton.button == 3)
        do_action(button->action_right);
     else if (ev.xbutton.button == 2)
        do_action(button->action_middle);
     else
        do_action(button->action_left);
   }
   else if (ev.type == Expose)
      button->exposed = TRUE;
}

int isEventButton(Button *button, XEvent ev)
{
   if (ev.xbutton.subwindow == button->win)
      return TRUE;
   else if (ev.xcrossing.subwindow == button->win)
      return TRUE;
   else if (ev.xbutton.window == button->win)
      return TRUE;
   else if (ev.xcrossing.window == button->win)
      return TRUE;
   return FALSE;
}

void FreeButton(Button *button)
{
   if (!button)
      return;
   FreeImageDat(button->normal);
   FreeImageDat(button->clicked);
   FreeImageDat(button->over);
   button->normal = NULL;
   free(button);
   button = NULL;
}

void CreateMultiButton(MultiButton *mbutton, Window parent)
{
  struct _multibutton *cur;
  if (!mbutton)
     return;
  if (!mbutton->first || !mbutton->last)
     return;
  cur = mbutton->last;
  while (cur != mbutton->first) {
     if (deferred)
        CreateButton(cur->button, parent);
     HideButton(cur->button);
     cur = cur->prev;
  }
  if (deferred)
     CreateButton(mbutton->first->button, parent);
  UnhideButton(mbutton->first->button);
  mbutton->visible = TRUE;
  mbutton->current = mbutton->first;
  *(mbutton->var) = 0;
  if (!mbutton->state)
     HideMultiButton(mbutton);
}

void DefaultsMultiButton(MultiButton *mbutton)
{
   mbutton->number = 0;
   mbutton->state = 1;
   mbutton->first = mbutton->last = mbutton->current = NULL;
}

void ConfigMultiButton(MultiButton *mbutton, Window parent, FILE *file, int *var)
{
   struct _multibutton *cur = NULL;
   char st[256];
   char s1[128], s2[128];
   int end = FALSE;
   DefaultsMultiButton(mbutton);
   mbutton->var = var;
   while (!end) {
      if (!fgets(st, 256, file)) {
	 fprintf(stderr,"Unexpected EOF in config file\n");
	 exit(-1);
      }
      if ((st[0] != '#') && (st[0] != '\n') && (st[0] != '\0')) {
	 splitstring(s1, s2, st);
	 if (!strncmp("END", s1, 3))
	    end = TRUE;
	 else if (!strncmp("NUMBER", s1, 6))
	    mbutton->number = atoi(s2);
	 else if (!strncmp("VISIBLE", s1, 7))
	    mbutton->state = parsebool(s2);
	 else if (!strncmp("BEGIN", s1, 5)) {
	    if (!mbutton->first) {
	       mbutton->first = malloc(sizeof(struct _multibutton));
	       cur = mbutton->first;
	       cur->prev = NULL;
	       cur->next = NULL;
	    }
	    else {
	       cur->next = malloc(sizeof(struct _multibutton));
	       cur->next->prev = cur;
	       cur = cur->next;
	       cur->next = NULL;
	    }
	    cur->button = malloc(sizeof(Button));
	    ConfigButton(cur->button, parent, file);
	    mbutton->last = cur;
	 }
      }
   }
   if (!deferred)
      CreateMultiButton(mbutton, parent);
}

void SetVarMultiButton(MultiButton *mbutton, int *var)
{
   mbutton->var = var;
}

void SetMultiButton(MultiButton *mbutton, int number)
{
   struct _multibutton *cur;
   int counter = 1;
   if (!mbutton)
      return;
   if (number > mbutton->number)
      return;
   cur = mbutton->first;
   HideButton(mbutton->current->button);
   while (counter != number && cur) {
      cur = cur->next;
      if (cur == mbutton->last)
         break;
   }
   mbutton->current = cur;
   *mbutton->var = number - 1;
   UnhideButton(mbutton->current->button);
}

void EventMultiButton(MultiButton *mbutton, XEvent ev)
{
   if (!mbutton || !mbutton->current)
      return;
   if (ev.type != ButtonRelease)
      EventButton(mbutton->current->button, ev);
   if (ev.type == ButtonRelease) {
      (*mbutton->var)++;
      if (*mbutton->var == mbutton->number)
	 *mbutton->var = 0;
      EventButton(mbutton->current->button, ev);
      HideButton(mbutton->current->button);
      if (mbutton->current == mbutton->last)
	 mbutton->current = mbutton->first;
      else
	 mbutton->current = mbutton->current->next;
      UnhideButton(mbutton->current->button);
   }
}

int isEventMultiButton(MultiButton *mbutton, XEvent ev)
{
   if (isEventButton(mbutton->current->button, ev))
      return TRUE;
   return FALSE;
}

void FreeMultiButton(MultiButton *mbutton)
{
   if (!mbutton || !mbutton->current)
      return;
   mbutton->current = mbutton->first;
   while (mbutton->current) {
      FreeButton(mbutton->current->button);
      mbutton->current = mbutton->current->next;
      if (mbutton->current) {
         free(mbutton->current->prev);
         mbutton->current->prev = NULL;
      }
   }
   free(mbutton->last);
   mbutton->last = NULL;
   mbutton->first = NULL;
   mbutton->current = NULL;
   free(mbutton);
   mbutton = NULL;
}
