#include "../emusic.h"

void HideContainer(Container *cont)
{
   if (cont->visible) {
      XUnmapWindow(disp, cont->win);
      cont->visible = FALSE;
   }
}

void UnhideContainer(Container *cont)
{
   if (!cont->visible) {
      XMapWindow(disp, cont->win);
      cont->visible = TRUE;
   }
}

void CreateContainer(Container *cont, Window parent)
{
   if (!cont)
      return;
   cont->parent = parent;
   if (!cont->parent)
      cont->parent = root;
   cont->win = CreateWin(cont->parent, cont->x, cont->y, cont->w, cont->h);
   XMapWindow(disp, cont->win);
   XSetWindowBackground(disp, cont->win, 0);
}

void DefaultsContainer(Container *cont)
{
   cont->x = cont->y = cont->w = cont->h = 0;
   cont->visible = TRUE;
   cont->state = 1;
   cont->exposed = 0;
   cont->action_left = cont->action_middle = cont->action_right = 0;
}

void ConfigContainer(Container *cont, Window parent, FILE *file)
{
   char st[256];
   char s1[128], s2[128];
   int end = FALSE;
   DefaultsContainer(cont);
   cont->parent = parent;
   while (!end) {
      if (!fgets(st, 256, file)) {
         fprintf(stderr,"Unexpected EOF in config file\n");
         exit(-1);
      }
      if ((st[0] != '#') && (st[0] != '\n') && (st[0] != '\0')) {
         splitstring(s1, s2, st);
	 if (!strncmp("END", s1, 3))
	    end = TRUE;
	 else if (!strncmp("LOCATION", s1, 8))
	    getcoords(s2, &cont->x, &cont->y);
	 else if (!strncmp("SIZE", s1, 4))
	    getcoords(s2, &cont->w, &cont->h);
	 else if (!strncmp("VISIBLE", s1, 7))
	    cont->state = parsebool(s2);
	 else if (!strncmp("ACTION_R", s1, 8))
	    cont->action_right = parse_action(s2);
	 else if (!strncmp("ACTION_M", s1, 8))
	    cont->action_middle = parse_action(s2);
	 else if (!strncmp("ACTION_L", s1, 8))
	    cont->action_left = parse_action(s2);
      }
   }
   if (!deferred)
      CreateContainer(cont, parent);
}

void EventContainer(Container *cont, XEvent ev)
{
   if (ev.type != ButtonRelease)
      return;
   if (ev.xbutton.button == 3 && !cont->action_right)
      ev.xbutton.button = 1;
   else if (ev.xbutton.button == 2 && !cont->action_middle)
      ev.xbutton.button = 1;
   if (ev.xbutton.button == 3)
      do_action(cont->action_right);
   else if (ev.xbutton.button == 2)
      do_action(cont->action_middle);
   else
      do_action(cont->action_left);
}

int isEventContainer(Container *cont, XEvent ev)
{
   if (ev.xbutton.subwindow == cont->win)
      return TRUE;
   if (ev.xany.window == cont->win)
      return TRUE;
   return FALSE;
}

void FreeContainer(Container *cont)
{
   if (!cont)
      return;
   free(cont);
   cont = NULL;
}
