#include "../emusic.h"

void HideFileBox(FileBox *filebox)
{
   if (filebox->visible) {
      XUnmapWindow(disp, filebox->win);
      filebox->visible = FALSE;
   }
}

void UnhideFileBox(FileBox *filebox)
{
   if (!filebox->visible) {
      XMapWindow(disp, filebox->win);
      filebox->visible = TRUE;
   }
}

void DefaultsFileBox(FileBox *filebox)
{
   filebox->x = filebox->y = filebox->w = filebox->h = 0;
   filebox->back = NULL;
   filebox->ok = filebox->cancel = NULL;
   filebox->recursive = NULL;
   filebox->returnpath = NULL;
   filebox->filelist = NULL;
   filebox->done = filebox->exposed = FALSE;
   filebox->visible = TRUE;
   filebox->cdmode = FALSE;
   filebox->recurse = 0;
   filebox->okdone = FALSE;
}

void CreateFileBox(FileBox *filebox)
{
   char temp[256];
   getcwd(temp, 256);
   SetTextTextBox(filebox->returnpath, temp);
   UpdateTextBox(filebox->returnpath);
   if (!filebox->state)
      HideFileBox(filebox);
}

void ConfigFileBox(FileBox *filebox, FILE *file)
{
   char st[256];
   char s1[128], s2[128];
   int end = FALSE;
   DefaultsFileBox(filebox);
   while (!end) {
      if (!fgets(st, 256, file)) {
	 fprintf(stderr,"Unexpected EOF in config file\n");
	 exit(-1);
      }
      if ((st[0] != '#') && (st[0] != '\n') && (st[0] != '\0')) {
	 splitstring(s1, s2, st);
	 if (!strncmp("END", s1, 3))
	    end = TRUE;
	 else if (!strncmp("VISIBLE", s1, 7)) {
	    filebox->state = parsebool(s2);
	    filebox->deferred = deferred = !filebox->state;
	 }
	 else if (!strncmp("BEGIN", s1, 5)) {
	    upstr(s2);
	    if (!strncmp("BACKGROUND", s2, 10)) {
	       filebox->back = malloc(sizeof(Background));
	       ConfigBack(filebox->back, file);
	       filebox->win = filebox->back->win;
	    }
	    else if (!strncmp("OK", s2, 2)) {
	       filebox->ok = malloc(sizeof(Button));
	       ConfigButton(filebox->ok, filebox->win, file);
	    }
	    else if (!strncmp("CANCEL", s2, 6)) {
	       filebox->cancel = malloc(sizeof(Button));
	       ConfigButton(filebox->cancel, filebox->win, file);
	    }
	    else if (!strncmp("RECURSIVE", s2, 9)) {
	       filebox->recursive = malloc(sizeof(MultiButton));
	       ConfigMultiButton(filebox->recursive, filebox->win, file, &filebox->recurse);
	    }
	    else if (!strncmp("RETURNPATH", s2, 10)) {
	       filebox->returnpath = malloc(sizeof(TextBox));
	       ConfigTextBox(filebox->returnpath, filebox->win, file);
	    }
	    else if (!strncmp("FILELIST", s2, 8)) {
	       filebox->filelist = malloc(sizeof(ListBox));
	       ConfigListBox(filebox->filelist, filebox->win, file);
	    }
	 }
      }
   }
   if (!deferred)
      CreateFileBox(filebox);
   else
      filebox->visible = FALSE;
   deferred = FALSE;
}

void Handle_Keys(FileBox *filebox, XEvent ev)
{
   KeySym keysym;
   char buffer[10];
   XComposeStatus compose;
   char *temp = NULL, temp2[128];
   int count;
   char returnpath[256];
   strcpy(returnpath,ReturnTextTextBox(filebox->returnpath));
   count = XLookupString(&ev.xkey, buffer, 10, &keysym, &compose);
   buffer[count] = '\0';
   if ((keysym == XK_Return) || (keysym == XK_KP_Enter) || (keysym == XK_Linefeed)) {
     if (!chdir(returnpath))
         GetCurrentDirectory(filebox);
     else {
        temp = splitdir(returnpath);
        strcpy(temp2,temp);
        if (strlen(returnpath) == 0)
	   strcpy(returnpath,"/"); 
        chdir(returnpath);
        GetCurrentDirectory(filebox);
        strcat(returnpath,"/");
	strcat(returnpath,temp2);
	} 
     } 
   else if (((keysym >= XK_KP_Space) && (keysym <= XK_KP_9)) || 
            ((keysym >= XK_space) && (keysym <= XK_asciitilde)))
     strcat(returnpath, buffer);
   else if ((keysym == XK_BackSpace) || (keysym == XK_Delete)) {
     if (strlen(returnpath) > 0) 
	 returnpath[(strlen(returnpath)-1)] = '\0'; 
   }
   SetTextTextBox(filebox->returnpath, returnpath);
   UpdateTextBox(filebox->returnpath);
}

void EventFileBox(FileBox *filebox, XEvent ev)
{
   if (isEventButton(filebox->ok, ev))
      EventButton(filebox->ok, ev);
   else if (isEventButton(filebox->cancel, ev))
      EventButton(filebox->cancel, ev);
   else if (isEventListBox(filebox->filelist, ev))
      EventListBox(filebox->filelist, ev);
   else if (isEventMultiButton(filebox->recursive, ev))
      EventMultiButton(filebox->recursive, ev);
   else if (isEventTextBox(filebox->returnpath, ev)) {
      if (ev.type == KeyPress)
	 Handle_Keys(filebox, ev);
      else
	 EventTextBox(filebox->returnpath, ev);
   }
   else if (ev.xany.window == filebox->win)
      UpdateTextBox(filebox->returnpath);
}

void Display_File(FileBox *filebox)
{
   if (filebox->exposed) {
      UpdateTextBox(filebox->returnpath);
      RedrawListBoxText(filebox->filelist);
      filebox->exposed = FALSE;
   }
} 

void KillFileBox(FileBox *filebox)
{
   if (!filebox->visible)
      return;
   clearplaylist(filebox->filelist->list);
   free(filebox->filelist->list);
   filebox->filelist->list = NULL;
   HideFileBox(filebox);
   filebox->callback();
}

void GetCurrentDirectory(FileBox *filebox)
{
   DIR *dir;
   struct dirent *entry;
   struct stat statbuf;
   struct playlist_item *cur;
   char cwd[256];
   int i;
   if (filebox->cdmode)
      makelistofnums(filebox->filelist->list, getnumtracks_cda());
   else {
      getcwd(cwd, 256);
      if (!(dir = opendir(cwd))) {
	 fprintf(stderr,"Can't read directory %s\n", cwd);
	 return;
      }
      clearplaylist(filebox->filelist->list);
      filebox->filelist->list->first = malloc(sizeof(struct playlist_item));
      cur = filebox->filelist->list->first;
      cur->prev = NULL;
      i = 1;
      while ((entry = readdir(dir))) {
	 stat(entry->d_name, &statbuf);
	 if (S_ISDIR(statbuf.st_mode))
	    cur->length = 1;
	 else
	    cur->length = 0;
	 if (strlen(cwd) == 1)
	    strcpy(cwd,"\0");
	 sprintf(cur->name,"%s", entry->d_name);
	 cur->pos = i;
	 i++;
	 cur->next = malloc(sizeof(struct playlist_item));
	 cur->next->prev = cur;
	 cur = cur->next;
      }
      cur = cur->prev;
      if (cur->next)
	 free(cur->next);
      cur->next = NULL;
      filebox->filelist->list->last = cur;
   }
   filebox->filelist->current_selection = filebox->filelist->start = 1;
   sort_list(filebox->filelist->list, 0);
   renumber_list(filebox->filelist->list, 0);
   SetPlaylistListBox(filebox->filelist, filebox->filelist->list);
   DisplayListBox(filebox->filelist);
}

void SelectCurrentFileFileBox(FileBox *filebox, int current)
{
   char returnpath[256];
   struct playlist_item *cur;
   if (current) {
      if (!filebox->filelist->list->first) 
	 return;
      strcpy(returnpath, ReturnTextTextBox(filebox->returnpath));
      cur = filebox->filelist->list->first;
      while (cur && (cur->pos != current))
	 cur = cur->next;
      if (!cur)
	 return;
      if (!filebox->cdmode) {
	 getcwd(returnpath, 256);
	 if (strlen(returnpath) > 1)
	    strcat(returnpath,"/");
	 strcat(returnpath, cur->name);
	 if ((cur->length == 1) && (current != 1)) {
	    chdir(returnpath);
	    GetCurrentDirectory(filebox);
	    getcwd(returnpath, 256);
	 }
      }
      else
	 strcpy(returnpath, cur->path);
   }
   else
      strcpy(returnpath,"/");
   SetTextTextBox(filebox->returnpath, returnpath);
   UpdateTextBox(filebox->returnpath);
}

void PopUpFileBox(FileBox *filebox)
{
   if (filebox->visible)
      return;
   if (filebox->deferred) {
      deferred = TRUE;
      CreateBack(filebox->back);
      filebox->win = filebox->back->win;
      CreateButton(filebox->ok, filebox->win);
      CreateButton(filebox->cancel, filebox->win);
      CreateMultiButton(filebox->recursive, filebox->win);
      CreateListBox(filebox->filelist, filebox->win);
      CreateTextBox(filebox->returnpath, filebox->win);
      filebox->deferred = FALSE;
      deferred = FALSE;
   }
   UnhideFileBox(filebox);
   SetSliderPosition(filebox->filelist->vertical, 0.01);
   SetSliderPosition(filebox->filelist->horizontal, 0.01);
   filebox->filelist->list = malloc(sizeof(PlayList));
   filebox->filelist->list->first = NULL;
   filebox->filelist->list->last = NULL;
   filebox->done = FALSE;
   filebox->filelist->current_selection = 1;
   filebox->filelist->start = 1;
   filebox->recurse = 0;
   SetMultiButton(filebox->recursive, 1);
   GetCurrentDirectory(filebox);
   SelectCurrentFileFileBox(filebox, 1);
}

void FreeFileBox(FileBox *filebox)
{
   HideFileBox(filebox);
   XSync(disp, False);
   clearplaylist(filebox->filelist->list);
   free(filebox->filelist->list);
   filebox->filelist->list = NULL;
   FreeBack(filebox->back);
   FreeButton(filebox->ok);
   FreeButton(filebox->cancel);
   FreeMultiButton(filebox->recursive);
   FreeTextBox(filebox->returnpath);
   FreeListBox(filebox->filelist);
   free(filebox);
   filebox = NULL;
}

void SetFileBoxCallback(FileBox *filebox, void *call)
{
   filebox->callback = call;
}

