#include "../emusic.h"

void HideListBox(ListBox *list)
{
   if (list->visible) {
      HideSlider(list->vertical);
      HideSlider(list->horizontal);
      XUnmapWindow(disp, list->win);
      list->visible = FALSE;
   }
}

void UnhideListBox(ListBox *list)
{
   if (!list->visible) {
      XMapWindow(disp, list->win);
      UnhideSlider(list->vertical);
      UnhideSlider(list->horizontal);
      list->visible = TRUE;
      DisplayListBox(list);
   }
}

void DefaultsListBox(ListBox *list)
{
   list->x = list->y = list->w = list->h = 0;
   list->back = NULL;
   list->state = 1;
   list->exposed = FALSE;
   list->fontheight = 1;
   list->font_info = NULL;
   list->items = malloc(sizeof(XTextItem));
   list->items->chars = NULL;
   strcpy(list->fontname,"fixed");
   list->vertical = list->horizontal = NULL;
   list->current_selection = list->last_selection = list->click_time = 0;
   list->list = NULL;
   list->start = 1;
   list->action = 0;
   list->visible = FALSE;
   list->use_fn = FALSE;
   list->fn = NULL;
}

void CreateListBox(ListBox *list, Window parent)
{
   list->parent = parent;
   if (!list->parent)
      list->parent = root;
   if (deferred) { 
      CreateSlider(list->vertical, list->parent);
      CreateSlider(list->horizontal, list->parent);
   }
   list->win = CreateWin(list->parent, list->x, list->y, list->w, list->h);
   XMapWindow(disp, list->win);
   if (list->back) {
      XSetWindowBackgroundPixmap(disp, list->win, list->back->pixmap);
      XShapeCombineMask(disp, list->win, ShapeBounding, 0, 0, list->back->mask, ShapeSet);
   }
   else {
      XImage *image;
      image = XGetImage(disp, list->parent, list->x +1, list->y +1, 1, 1, AllPlanes, ZPixmap);
      XSetWindowBackground(disp, list->win, XGetPixel(image, 0, 0));
   }
   XClearWindow(disp, list->win);
   RaiseSlider(list->vertical);
   RaiseSlider(list->horizontal);
   list->visible = TRUE;
   if ((list->fn = Fnlib_load_font(fnlib, list->fontname))) {
      list->use_fn = TRUE;
      list->fnstyle.mode = MODE_VERBATIM;
      list->fnstyle.orientation = FONT_TO_RIGHT;
      list->fnstyle.justification = 0;
      list->fnstyle.spacing = 0;
   } 
   else if ((list->font_info = XLoadQueryFont(disp, list->fontname)) == NULL) {
      fprintf(stderr,"The font %s could not be loaded.  Using fixed instead\n", list->fontname);
      if ((list->font_info = XLoadQueryFont(disp, "fixed")) == NULL) {
	 fprintf(stderr,"Couldn't load fixed either, quitting\n");
	 exit(-1);
      }
   }
   if (!list->use_fn) {
      list->items->font = list->font_info->fid;   
      list->fontheight = list->font_info->max_bounds.ascent + list->font_info->max_bounds.descent;
   }
   list->items->delta = 0;
   if (!list->state)
      HideListBox(list);
   Imlib_free_pixmap(imlib, list->back->pixmap);
   Imlib_free_pixmap(imlib, list->back->mask);
}

void ConfigListBox(ListBox *list, Window parent, FILE *file)
{
   char st[256];
   char s1[128], s2[128];
   int end = FALSE;
   DefaultsListBox(list);
   list->parent = parent;
   while (!end) {
      if (!fgets(st, 256, file)) {
	 fprintf(stderr,"Unexpected EOF in config file\n");
	 exit(-1);
      }
      if ((st[0] != '#') && (st[0] != '\n') && (st[0] != '\0')) {
	 splitstring(s1, s2, st);
	 if (!strncmp("END", s1, 3))
	    end = TRUE;
	 else if (!strncmp("BACKGROUND", s1, 10)) {
	    list->back = malloc(sizeof(ImageDat));
	    LoadImage(s2, list->back);
	    list->w = list->back->w;
	    list->h = list->back->h;
	 }
	 else if (!strncmp("LOCATION", s1, 8)) 
	    getcoords(s2, &list->x, &list->y);
	 else if (!strncmp("SIZE", s1, 4))
	    getcoords(s2, &list->w, &list->h);
	 else if (!strncmp("VISIBLE", s1, 7))
	    list->state = parsebool(s2);
	 else if (!strncmp("FONTNAME", s1, 8))
	    strcpy(list->fontname, s2);
	 else if (!strncmp("FONTHEIGHT", s1, 10))
            list->fontheight = atoi(s2);
	 else if (!strncmp("BEGIN", s1, 5)) {
	    upstr(s2);
	    if (!strncmp("VERTICAL", s2, 8)) {
	       list->vertical = malloc(sizeof(Slider));
	       ConfigSlider(list->vertical, parent, file);
	    }
	    else if (!strncmp("HORIZONTAL", s2, 10)) {
	       list->horizontal = malloc(sizeof(Slider));
	       ConfigSlider(list->horizontal, parent, file);
	    }
	 }
	 else if (!strncmp("ACTION", s1, 6))
	    list->action = parse_action(s2);
	 else fprintf(stderr,"Unknown line in ListBox config\n");
      }
   }
   if (!deferred)
      CreateListBox(list, parent);
}

void EventListBox(ListBox *list, XEvent ev)
{
   if (isEventSlider(list->vertical, ev)) { 
      EventSlider(list->vertical, ev);
      DisplayListBox(list);
   }
   else if (isEventSlider(list->horizontal, ev)) {
      EventSlider(list->horizontal, ev);
      DisplayListBox(list);
   }
   else if (ev.type != ButtonRelease)
      return;
   else {
      list->current_selection = ceil((double)(ev.xbutton.y) / (double)list->fontheight) + list->start - 1;
      RedrawListBoxText(list);
      if ((abs(list->click_time - tv.tv_sec) < 1) && (list->current_selection == list->last_selection)) {
	 list->last_selection = 0;
	 do_action(list->action);
      }
      else {
	 list->last_selection = list->current_selection;
	 list->click_time = tv.tv_sec;
      }
   }
}

int isEventListBox(ListBox *list, XEvent ev)
{
   if (isEventSlider(list->vertical, ev))
      return TRUE;
   if (isEventSlider(list->horizontal, ev))
      return TRUE;
   if (ev.xbutton.subwindow == list->win)
      return TRUE;
   if (ev.xany.window == list->win) 
      return TRUE;
   return FALSE;
}

void RedrawListBoxText(ListBox *list)
{
   struct playlist_item *cur;
   int max_items, y_pos = 0, i;
   if (!list->list)
      return;
   XClearWindow(disp, list->win);
   max_items = list->h / list->fontheight;
   if (!list->list->first) 
      return;
   else {
     cur = list->list->first;
     while (cur && (cur->pos != list->start)) {
        cur = cur->next;
     }
   }
   y_pos = list->fontheight - list->fontheight / 3;
   for (i = 1; i <= max_items; i++) {
      if (!cur) 
	 break;
      list->items->nchars = strlen(cur->name);
      list->items->chars = cur->name;
      if (cur->length == 1) {
	 list->items->nchars += 1;
	 strcat(list->items->chars, "/");
      }
      if (cur->pos != list->current_selection) {
	 if (!list->use_fn)
	    XDrawText(disp, list->win, fg, 0, y_pos, list->items, 1);
      }
      else {
	 XFillRectangle(disp, list->win, fg, 0, y_pos - list->fontheight + 2, list->w - list->vertical->w, list->fontheight);
         XDrawText(disp, list->win, bg, 0, y_pos, list->items, 1);
      }
      if (cur->length == 1)
	 list->items->chars[(list->items->nchars - 1)] = '\0';
      if (!cur->next)
	 break;
      else {
	 cur = cur->next;
	 y_pos += list->fontheight;
      }
   }
   if (!list->list->first)
      free(cur);
}

void DisplayListBox(ListBox *list)
{

   if (list->vertical->dragging) {
      if (list->list->last)
	 list->start = (((float)(list->vertical->slider_position - list->vertical->min) / (float)list->vertical->h * (float)list->list->last->pos));
      else
	 list->start = 1;
      if (list->start <= 0) 
	 list->start = 1;
      if (list->list->last && (list->start >= (list->list->last->pos - 5))) 
	 list->start = list->list->last->pos - 5;
   }
   if (list->horizontal->dragging) {
      list->items->delta = -(list->horizontal->slider_position - list->horizontal->min);
   }
   RedrawListBoxText(list);
}

void SetPlaylistListBox(ListBox *list, PlayList *plist)
{
   list->list = plist;
   if (list->visible)
      DisplayListBox(list);
}

void FreeListBox(ListBox *list)
{
   if (!list)
      return;
   FreeImageDat(list->back);   
   if (list->font_info)
      free(list->font_info);
   if (list->items) {
      free(list->items);
      list->items = NULL;
   }
   if (list->fn)
      Fnlib_free_font(fnlib, list->fn);
   FreeSlider(list->vertical);
   FreeSlider(list->horizontal);
   free(list);
   list = NULL;
}
