#include "../emusic.h"

void HideSlider(Slider *slider)
{
   if (slider->visible) {
      if (slider->arrows) {
         HideButton(slider->arrow1);
       	 HideButton(slider->arrow2);
      }
      HideButton(slider->button);
      XUnmapWindow(disp, slider->win);
      slider->visible = FALSE;
   }
}

void UnhideSlider(Slider *slider)
{
   if (!slider->visible) {
      XMapWindow(disp, slider->win);
      if (slider->arrows) {
	 UnhideButton(slider->arrow1);
	 UnhideButton(slider->arrow2);
      }
      UnhideButton(slider->button);
      slider->visible = TRUE;
   }
}

void RaiseSlider(Slider *slider)
{
   XRaiseWindow(disp, slider->win);
   if (slider->arrows) {
      XRaiseWindow(disp, slider->arrow1->win);
      XRaiseWindow(disp, slider->arrow2->win);
   }
   XRaiseWindow(disp, slider->button->win);
}

void DefaultsSlider(Slider *slider)
{
   slider->x = slider->y = slider->w = slider->h = 0;
   slider->button = slider->arrow1 = slider->arrow2 = NULL;
   slider->back = NULL;
   slider->arrows = FALSE;
   slider->percent = 0;
   slider->horizontal = TRUE;
   slider->slider_size = 0;
   slider->slider_position = 0;
   slider->min = 0;
   slider->max = 0;
   slider->state = 1;
   slider->dragging = FALSE;
}

void CreateSlider(Slider *slider, Window parent)
{
   slider->parent = parent;
   if (!slider->parent)
      slider->parent = root;
   if (deferred) {
     if (slider->arrows) {
	 CreateButton(slider->arrow1, parent);
         CreateButton(slider->arrow2, parent);
     }
     CreateButton(slider->button, parent);
   }
   slider->win = CreateWin(slider->parent, slider->x, slider->y, slider->w, slider->h);
   XMapWindow(disp, slider->win);
   XSetWindowBackgroundPixmap(disp, slider->win, slider->back->pixmap);
   XShapeCombineMask(disp, slider->win, ShapeBounding, 0, 0, slider->back->mask, ShapeSet);
   XClearWindow(disp, slider->win);
   slider->visible = TRUE;
   if (slider->horizontal)
      slider->slider_size = slider->button->w;
   else
      slider->slider_size = slider->button->h;
   if (slider->arrows) {
      XMoveWindow(disp, slider->arrow1->win, slider->x, slider->y);
      slider->arrow1->x = slider->x;
      slider->arrow1->y = slider->y;
      if (slider->horizontal) {
	 slider->min = slider->arrow1->w + slider->x;
	 slider->max = slider->w - slider->arrow2->w + slider->x;
	 XMoveWindow(disp, slider->arrow2->win, slider->max, slider->y);
	 slider->arrow2->x = slider->max;
	 slider->arrow2->y = slider->y;
      }
      else {
	 slider->min = slider->arrow1->h + slider->y;
	 slider->max = slider->h - slider->arrow2->h + slider->y;
	 XMoveWindow(disp, slider->arrow2->win, slider->x, slider->max);
	 slider->arrow2->x = slider->x;
	 slider->arrow2->y = slider->max;
      }
   }
   else {
     if (slider->horizontal) {
        slider->min = slider->x;
        slider->max = slider->w + slider->x;
     }
     else {
        slider->min = slider->y;
	slider->max = slider->h + slider->y;
     }
   }
   slider->max -= slider->slider_size;
   if (slider->horizontal) {
      XMoveWindow(disp, slider->button->win, slider->min, slider->y);
      slider->button->x = slider->min;
      slider->button->y = slider->y;
   }
   else {
      XMoveWindow(disp, slider->button->win, slider->x, slider->min);
      slider->button->x = slider->x;
      slider->button->y = slider->min;
   }
   XRaiseWindow(disp, slider->button->win);
   slider->percent = 0;
   slider->slider_position = slider->min;
   if (!slider->state)
      HideSlider(slider);
   Imlib_free_pixmap(imlib, slider->back->pixmap);
   Imlib_free_pixmap(imlib, slider->back->mask);
}

void ConfigSlider(Slider *slider, Window parent, FILE *file)
{
   char st[256];
   char s1[128], s2[128];
   int end = FALSE;
   DefaultsSlider(slider);
   slider->parent = parent;
   while (!end) {
      if (!fgets(st, 256, file)) {
	 fprintf(stderr,"Unexpected EOF in config file\n");
	 exit(-1);
      }
      if ((st[0] != '#') && (st[0] != '\n') && (st[0] != '\0')) {
	 splitstring(s1, s2, st);
	 if (!strncmp("END", s1, 3))
	    end = TRUE;
	 else if (!strncmp("BACKGROUND", s1, 10)) {
	    slider->back = malloc(sizeof(ImageDat));
	    LoadImage(s2, slider->back);
	    slider->w = slider->back->w;
	    slider->h = slider->back->h;
	 }
	 else if (!strncmp("ARROWS", s1, 6))
	    slider->arrows = parsebool(s2);
	 else if (!strncmp("HORIZONTAL", s1, 10)) 
	    slider->horizontal = parsebool(s2);
	 else if (!strncmp("LOCATION", s1, 8))
	    getcoords(s2, &slider->x, &slider->y);
	 else if (!strncmp("VISIBLE", s1, 7))
	    slider->state = parsebool(s2);
	 else if (!strncmp("BEGIN", s1, 5)) {
	   upstr(s2);
	   if (!strncmp("BUTTON", s2, 6)) {
	      slider->button = malloc(sizeof(Button));
	      ConfigButton(slider->button, parent, file);
	   }
	   else if (!strncmp("ARROW1", s2, 6)) {
	      slider->arrow1 = malloc(sizeof(Button));
	      ConfigButton(slider->arrow1, parent, file);
	   }
	   else if (!strncmp("ARROW2", s2, 6)) {
	      slider->arrow2 = malloc(sizeof(Button));
	      ConfigButton(slider->arrow2, parent, file);
	   }
	 }
	 else
	    fprintf(stderr,"Unknown line in Slider config\n");
      }
   }
   if (!deferred)
      CreateSlider(slider, parent);
}

void EventSlider(Slider *slider, XEvent ev)
{
   float per;
   static int offset = 0;
   if (ev.type == ButtonPress) {
      if (ev.xbutton.window == slider->button->win) { 
	 EventButton(slider->button, ev);
	 slider->dragging = -1;
	 if (slider->horizontal) {
            offset = ev.xbutton.x_root;
	    slider->slider_position = ev.xbutton.x_root - offset;
	 }
	 else {
	    offset = ev.xbutton.y_root;
	    slider->slider_position = ev.xbutton.y_root - offset;
	 }
      }
      else if (slider->arrows) {
	 if (ev.xbutton.subwindow == slider->arrow1->win) {
	    EventButton(slider->arrow1, ev);
	    per = GetSliderPosition(slider);
	    if (per >= 2)
	       per -= 2.0;
	    SetSliderPosition(slider, per);
	 }
	 else if (ev.xbutton.subwindow == slider->arrow2->win) {
	    EventButton(slider->arrow2, ev);
	    per = GetSliderPosition(slider);
	    if (per <= 98)
	       per += 2.0;
	    SetSliderPosition(slider, per);
	 }
      }
   }
   else if (ev.type == MotionNotify) {
      if (slider->horizontal) {
	 if ((slider->dragging != 0) && (slider->dragging != ev.xmotion.x_root)) {
	    slider->dragging = ev.xmotion.x_root;
	    slider->slider_position = ev.xmotion.x_root - offset + slider->min;
	    UpdateSlider(slider);
	 }
      }
      else {
	 if ((slider->dragging != 0) && (slider->dragging != ev.xmotion.y_root)) {
            slider->dragging = ev.xmotion.y_root;
	    slider->slider_position = ev.xmotion.y_root - offset + slider->min;
	    UpdateSlider(slider);
	 }
      }
   }
   else if (ev.type == ButtonRelease) {
      slider->dragging = FALSE;
      EventButton(slider->button, ev);
   }
}

int isEventSlider(Slider *slider, XEvent ev)
{
   if (isEventButton(slider->button, ev))
      return TRUE;
   if (slider->arrows) {
      if (isEventButton(slider->arrow1, ev))
	 return TRUE;
      if (isEventButton(slider->arrow2, ev))
	 return TRUE;
   }
   if (ev.xbutton.subwindow == slider->win)
      return TRUE;
   if (ev.xany.window == slider->win)
      return TRUE;
   return FALSE;
}

void UpdateSlider(Slider *slider)
{
   if (slider->slider_position < slider->min)
      slider->slider_position = slider->min;
   if (slider->slider_position > slider->max)
      slider->slider_position = slider->max;
   if (slider->horizontal) {
      XMoveWindow(disp, slider->button->win, slider->slider_position, slider->y);
      slider->button->x = slider->slider_position;
   }
   else {
      XMoveWindow(disp, slider->button->win, slider->x, slider->slider_position);
      slider->button->y = slider->slider_position;
   }
}

float GetSliderPosition(Slider *slider)
{
   int total;
   total = slider->max - slider->min;
   slider->percent = (float)(slider->slider_position - slider->min) / (float)total;
   return slider->percent;
}

void SetSliderPosition(Slider *slider, float percent)
{
   int total;
   if (percent == slider->percent)
      return;
   total = slider->max - slider->min;
   slider->slider_position = total * percent + slider->min;
   slider->percent = percent;
   UpdateSlider(slider);
}

void FreeSlider(Slider *slider)
{
   if (!slider)
      return;
   FreeImageDat(slider->back);
   FreeButton(slider->button);
   FreeButton(slider->arrow1);
   FreeButton(slider->arrow2);
   free(slider);
   slider = NULL;
}
