#include "../emusic.h"

void HideTextBox(TextBox *box)
{
   if (box->visible) {
      XUnmapWindow(disp, box->win);
      box->visible = FALSE;
   }
}

void UnhideTextBox(TextBox *box)
{
   if (!box->visible) {
      XMapWindow(disp, box->win);
      box->visible = TRUE;
   }
}

void DefaultsTextBox(TextBox *box)
{
   box->x = box->y = box->w = box->h = 0;
   box->back = NULL;
   box->state = 1;
   box->exposed = 0;
   box->interactive = 0;
   box->visible = TRUE;
   box->action_left = box->action_middle = box->action_right = 0;
   box->fontheight = 0;
   box->font_info = NULL;
   box->items = malloc(sizeof(XTextItem));
   box->items->chars = malloc(1);
   box->items->delta = 0;
   box->name_x = 5;
   box->name_width = 0;
   box->selected = FALSE;
   box->bounce = FALSE;
   box->scroll = FALSE;
   box->scroll_time = 0;
   strcpy(box->fontname, "fixed");
   box->use_fn = FALSE;
   box->fn = NULL;
}

void CreateTextBox(TextBox *box, Window parent)
{
   char text[20];
   XImage *image;
   box->parent = parent;
   if ((box->fn = Fnlib_load_font(fnlib, box->fontname))) {
      box->use_fn = TRUE;
      box->fnstyle.mode = MODE_VERBATIM;
      box->fnstyle.orientation = FONT_TO_RIGHT;
      box->fnstyle.justification = 0;
      box->fnstyle.spacing = 0;
   }
   else if ((box->font_info = XLoadQueryFont(disp, box->fontname)) == NULL) {
      fprintf(stderr,"The font %s could not be loaded.  Using fixed instead\n",box->fontname);
      if ((box->font_info = XLoadQueryFont(disp,"fixed")) == NULL) {
         fprintf(stderr,"Couldn't load fixed either, quitting\n");
         exit(-1);
      }
   }
   if (!box->use_fn) {
      box->items->font = box->font_info->fid;
      box->fontheight = box->font_info->max_bounds.ascent + box->font_info->max_bounds.descent;
      box->h = box->fontheight * 1.5;
   }
   if (!box->parent)
      box->parent = root;
   box->win = CreateWin(box->parent, box->x, box->y, box->w, box->h);
   XMapWindow(disp, box->win);
   XSync(disp, False);
/*   if (box->back) {
      XSetWindowBackgroundPixmap(disp, box->win, box->back->pixmap);
      XShapeCombineMask(disp, box->win, ShapeBounding, 0, 0, box->back->mask, ShapeSet);
   }  
   else */ {  
      image = XGetImage(disp, box->parent, box->x +1, box->y +1, 1, 1, AllPlanes, ZPixmap);
      XSetWindowBackground(disp, box->win, XGetPixel(image, 0, 0));
   } 
   XClearWindow(disp, box->win);
   if (!box->state)
      HideTextBox(box);
   sprintf(text,"eMusic %s",VERSION);
   SetTextTextBox(box, text);
   if (box->interactive)
      box->scroll = FALSE;
   if (box->back) {
      Imlib_free_pixmap(imlib, box->back->pixmap);
      Imlib_free_pixmap(imlib, box->back->mask);
   }
}

void ConfigTextBox(TextBox *box, Window parent, FILE *file)
{
   char st[256];
   char s1[128], s2[128];
   int end = FALSE;
   DefaultsTextBox(box);
   box->parent = parent;
   while (!end) {
      if (!fgets(st, 256, file)) {
	 fprintf(stderr,"Unexpected EOF in config file\n");
	 exit(-1);
      }
      if ((st[0] != '#') && (st[0] != '\n') && (st[0] != '\0')) {
	 splitstring(s1, s2, st);
	 if (!strncmp("END", s1, 3))
	    end = TRUE;
	 else if (!strncmp("BACKGROUND", s1, 10)) {
	    box->back = malloc(sizeof(ImageDat));
	    LoadImage(s2, box->back);
	    box->w = box->back->w;
	    box->h = box->back->h;
	 }
	 else if (!strncmp("SIZE", s1, 4)) {
	    if (box->back)
	       free(box->back);
	    box->back = NULL;
	    getcoords(s2, &box->w, &box->h);
	 }
	 else if (!strncmp("LOCATION", s1, 8))
	    getcoords(s2, &box->x, &box->y);
	 else if (!strncmp("INTERACTIVE", s1, 11)) 
	    box->interactive = parsebool(s2);
	 else if (!strncmp("VISIBLE", s1, 7))
	    box->state = parsebool(s2);
	 else if (!strncmp("FONTNAME", s1, 8))
	    strcpy(box->fontname, s2);
	 else if (!strncmp("FONTHEIGHT", s1, 10))
            box->fontheight = atoi(s2);
	 else if (!strncmp("SCROLL", s1, 6))
	    box->scroll = parsebool(s2);
	 else if (!strncmp("ACTION_R", s1, 8))
	    box->action_right = parse_action(s2);
	 else if (!strncmp("ACTION_M", s1, 8))
	    box->action_middle = parse_action(s2);
	 else if (!strncmp("ACTION_L", s1, 8))
	    box->action_left = parse_action(s2);
	 else
	    fprintf(stderr,"Unknown line in TextBox config\n");
      }
   }
   if (!deferred)
      CreateTextBox(box, parent);
}

void EventTextBox(TextBox *box, XEvent ev)
{
   if (ev.type == Expose)
      box->exposed = TRUE;
   else if (ev.type == ButtonRelease) {
      box->selected = TRUE;
      if ((ev.xbutton.button == 3) && (!box->action_right))
         ev.xbutton.button = 1;
      else if ((ev.xbutton.button == 2) && (!box->action_middle))
         ev.xbutton.button = 1;
      if (ev.xbutton.button == 3)
         do_action(box->action_right);
      else if (ev.xbutton.button == 2)
         do_action(box->action_middle);
      else
         do_action(box->action_left);
      UpdateTextBox(box);
   }
   else if (ev.type == KeyPress) 
      ;
}

int isEventTextBox(TextBox *box, XEvent ev)
{
   if (ev.xbutton.subwindow == box->win)
      return TRUE;
   else if (ev.xany.window == box->win)
      return TRUE;
   return FALSE;
}

void SetTextTextBox(TextBox *box, char *text)
{
   if (text == NULL) {
      text = malloc(20);
      sprintf(text,"eMusic %s",VERSION);
   }
   if (!box->items)
      box->items = malloc(sizeof(XTextItem));
   if (!strcmp(text, box->items->chars))
      return;
   box->items->nchars = strlen(text);
   box->items->chars = realloc(box->items->chars, strlen(text) + 1);
   strcpy(box->items->chars, text);
   if (!box->use_fn)
      box->name_width = XTextWidth(box->font_info, text, strlen(text));
   else 
      box->name_width = strlen(text) * 5;
   box->drawn = FALSE;
}

char *ReturnTextTextBox(TextBox *box)
{
   return box->items->chars;
}

void UpdateTextBox(TextBox *box)
{
   int x = 0, y = 0, rx = 0, ry = 0;
   static int old_x = -1;
   Window root_ret, child_ret;
   unsigned int mask_ret;
   if (!box || !box->items->chars)
      return;
   if (!box->interactive) {
      if (abs(tv.tv_usec - box->scroll_time) < 100000)
         return;
      box->scroll_time = tv.tv_usec;
   }
   if ((box->name_width <= box->w) || (!box->scroll)) {
      if (box->name_width > box->w && box->interactive)
	 box->name_x = box->w - box->name_width - 5;
      else
         box->name_x = box->items->delta = 0;
   }
   else if (box->bounce) {
      if (box->name_x > 0)
	 box->bounce = FALSE;
      else 
	 box->name_x += 5;
   }
   else {
      if ((box->name_x + box->name_width) < box->w)
	 box->bounce = TRUE;
      else
	 box->name_x -= 5;
   }
   if (box->drawn && (box->name_x == old_x))
      return;
   old_x = box->name_x;
   box->drawn = TRUE;
   XClearWindow(disp, box->win);
   if (box->use_fn)
      Fnlib_draw(fnlib, box->fn, box->win, 0, box->name_x, 0, box->name_width - box->name_x, box->h, 1, 0, box->fontheight, &box->fnstyle , box->items->chars);
   else 
      XDrawText(disp, box->win, fg, box->name_x, box->fontheight - box->fontheight / 3, box->items, 1);
   if (!box->interactive)
      return;
   if (box->selected) 
      if (!XQueryPointer(disp, box->win, &root_ret, &child_ret, &rx, &ry, &x, &y, &mask_ret)) 
	 box->selected = FALSE;
   if (box->selected) {
      if (!box->use_fn)
	 x = XTextWidth(box->font_info, box->items->chars, strlen(box->items->chars)) + 2;
      else
	 x = strlen(box->items->chars) * 5 + 2; /* FIXME this isn't always 5 */
      XDrawLine(disp, box->win, fg, x, 0, x, box->fontheight);
   }
}

void FreeTextBox(TextBox *text)
{
   if (!text)
      return;
   FreeImageDat(text->back);
   if (text->font_info)
      free(text->font_info);
   if (text->items) {
      if (text->items->chars)
         free(text->items->chars);
      free(text->items);
   }
   if (text->fn)
      Fnlib_free_font(fnlib, text->fn);
   free(text);
   text = NULL;
}
