#include "../emusic.h"

void RaiseLowerWidget(Widget *widget)
{
   if (widget->raised) {
      XLowerWindow(disp, widget->win);
      widget->raised = FALSE;
   }
   else {
      XRaiseWindow(disp, widget->win);
      widget->raised = TRUE;
   }
}

void MoveWidget(Widget *widget, Widget *offset, int x, int y)
{
   if (!widget->shaped)
      return;
   if (offset) {
      x += widget->offset_x;
      y += widget->offset_y;
   }
   XMoveWindow(disp, widget->win, x, y);
   widget->x = x;
   widget->y = y;
}

void HideWidget(Widget *widget)
{
   if (widget->visible) {
      XUnmapWindow(disp, widget->win);
      widget->visible = FALSE;
   }
}
void UnhideWidget(Widget *widget)
{
  if (!widget->visible) {
     XMapWindow(disp, widget->win);
     widget->visible = TRUE;
  }
}
void CreateWidget(Widget *widget)
{
   struct _widget *cur;
   if (!widget)
      return;
   if (!widget->first)
      return;
   deferred = TRUE;
   cur = widget->first;
   while (cur) {
      switch (cur->type) {
      case BACKGROUND: CreateBack(cur->widget);
	               widget->win = ((Background *)cur->widget)->win; 
		       MoveWidget(widget, MainWin, MainWin->x, MainWin->y); 
		       break;
      case BUTTON: CreateButton(cur->widget, widget->win); break;
      case MULTIBUTTON: CreateMultiButton(cur->widget, widget->win); break;
      case SLIDER: CreateSlider(cur->widget, widget->win); break;
      case TEXTBOX: CreateTextBox(cur->widget, widget->win); break;
      case LISTBOX: CreateListBox(cur->widget, widget->win); listbox = cur->widget; break;
      case CONTAINER: CreateContainer(cur->widget, widget->win); break;
      default: break;
      }
      cur = cur->next;
   }
   widget->deferred = FALSE;
   deferred = FALSE;
   widget->visible = TRUE;
}

void FreeWidget(Widget *widget) 
{
   struct _widget *cur, *temp;
   if (!widget)
      return;
   if (!widget->first)
      return;
   HideWidget(widget);
   XSync(disp, True);
   cur = widget->first;
   while (cur) {
      switch (cur->type) {
      case BACKGROUND: FreeBack(cur->widget); break;
      case BUTTON: FreeButton(cur->widget); break;
      case MULTIBUTTON: FreeMultiButton(cur->widget); break;
      case SLIDER: FreeSlider(cur->widget); break;
      case TEXTBOX: FreeTextBox(cur->widget); break;
      case LISTBOX: FreeListBox(cur->widget); break;
      case CONTAINER: FreeContainer(cur->widget);
      default: break;
      }
      temp = cur;
      cur = cur->next;
      free(temp);
      temp = NULL;
   }
   free(widget);
   widget = NULL;
}

int FindEventWidget(Widget *widget, XEvent ev)
{
   struct _widget *cur;
   int found = FALSE;
   if (!widget)
      return FALSE;
   if (!widget->first)
      return FALSE;
   cur = widget->first;
   while (cur && !found) {
      switch (cur->type) {
      case BUTTON: if (isEventButton(cur->widget, ev)) {
	               EventButton(cur->widget, ev);
                       found = TRUE;
                   }
                   break;
      case MULTIBUTTON: if (isEventMultiButton(cur->widget, ev)) {
	                    EventMultiButton(cur->widget, ev);
			    found = TRUE;
                        }
                        break;
      case SLIDER: if (isEventSlider(cur->widget, ev)) {
	               EventSlider(cur->widget, ev);
		       found = TRUE;
                   }
                   break;
      case TEXTBOX: if (isEventTextBox(cur->widget, ev)) {
	               found = TRUE;
	               EventTextBox(cur->widget, ev);
                    }
                    break;
      case LISTBOX: if (isEventListBox(cur->widget, ev)) {
	               EventListBox(cur->widget, ev);
                       found = TRUE;
                    }
                    break;
      case FILEBOX: if (ev.xany.window == ((FileBox *)cur->widget)->win) {
	               EventFileBox(cur->widget, ev);
		       found = TRUE;
                    }
                    break;
      case CONTAINER: if (isEventContainer(cur->widget, ev)) {
      	 	       EventContainer(cur->widget, ev);
      	 	       found = TRUE;
      	 	      }
      	 	    break;
      default: break;
      }
      cur = cur->next;
   }
   return found;
}
