#include "mpegsound.h"
#include "splay.h"
#include <string.h>
#include <sys/stat.h>
#include <math.h>
#include "../common.h"

Soundinputstream *loader = NULL;
Rawtoemusic	 *player = NULL;
Mpegtoraw	 *server = NULL;

int pcmperframe, frequency, maxframes;

extern int end_of_song;
extern int end_song;
extern int stop_song;
extern pthread_t player_thread;
extern pthread_mutex_t player_lock;

void seek_to_frame(int second)
{
   int frame;
   frame = second * frequency / pcmperframe;
   server->clearbuffer();
   server->setframe(frame);
}

void *mp3_thread(void *)
{
   pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
   pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
   while(1) {
      pthread_mutex_lock(&player_lock);
      if (!server->run(5) && (server->geterrorcode() == SOUND_ERROR_FINISH)) {
         end_song = 1;
	 server->clearbuffer();
	 pthread_mutex_unlock(&player_lock);
	 while (1)
	    pthread_testcancel();
      }
      else {
         pthread_testcancel();
	 pthread_mutex_unlock(&player_lock);
      }
   }
}

int play(int start, char *path)
{
   int err;
   end_of_song = 0;
   loader = Soundinputstream::hopen(path, &err);
   if (!loader)
      return 0;
   player = new Rawtoemusic;
   player->initialize(NULL);
   server = new Mpegtoraw(loader, player);
   server->initialize(path);
   server->setforcetomono(0);
   server->run(-1);
   frequency = server->getfrequency();
   pcmperframe = server->getpcmperframe();
   if (start)
      server->setframe(start * frequency / pcmperframe);
   pthread_create(&player_thread, NULL, mp3_thread, NULL);
   return 1;
}

void pause_it(void)
{
}

void stop(int playing)
{
   pthread_cancel(player_thread);
   usleep(100);
   if (!server)
      return;
   server->clearbuffer();
   player->abort();
   player->resetsoundtype();
   delete loader;
   delete player;
   delete server;
   server = NULL;
}

int get_second(void)
{
   return server->getcurrentframe() * pcmperframe / frequency;
}


int get_length(char *path)
{
   float len;
   struct stat mp3stat;
   stat(path, &mp3stat);
   len = mp3stat.st_size / 418.12;
   len = (int)(len / 16) * 16;
   return (int)(rint(len/38.28));
}

void set_name(struct playlist_item *song)
{
   if (song->tag.genre == -1)
      strcpy(song->name, strrchr(song->path,'/') + 1);
   else
      strcpy(song->name, song->tag.song);
}

int is_file(char *path)
{
   char *temp;
   if (!path)
      return 0;
   if ((temp = strrchr(path, '.')) && (temp+3)) {
      if (!strcasecmp((temp+1),"mp3")) 
         return 1;
      if (!strcasecmp((temp+1),"mpg")) 
         return 1;
   }
   return 0;
}

void version_info(char *id, char *version, char *copyright)
{
   strcpy(id, "splay");
   strcpy(version, "0.8.2");
   strcpy(copyright, "1998 Woo-jae Jung");
}

void init(void)
{
}

void get_info(struct playlist_item *song)
{
   song->tag.genre = -1;
}

void config(char *left, char *right)
{
}
