
#include "mod.h"
#include <string.h>

#include "xmp.h"
static struct xmp_module_info mi;
static int mod_length, tpo, bpm;
static double seconds;

static int mod_playing = FALSE;
static int mod_init = FALSE;

static void player_callback (unsigned long i)
{
    unsigned long msg = i >> 4;
    static int row, pos, pat;

    switch (i & 0xf) {
    case XMP_ECHO_BPM:
	bpm = msg & 0xff;
	tpo = msg >> 8;
	break;
    case XMP_ECHO_ORD:
	pos = msg & 0xff;
	pat = msg >> 8;
	break;
    case XMP_ECHO_ROW:
	row = msg & 0xff;
	seconds += 60.0 * tpo / (24 * bpm);
	break;
    }
}

void seek_to_frame(int second)
{
   if (second > seconds)
      xmp_player_ctl(XMP_ORD_NEXT);
   if (second < seconds)
      xmp_player_ctl(XMP_ORD_PREV);
}


int play(int start_second, char *song)
{
    struct xmp_options opt;
    if (!song) {
	fprintf (stderr, "No playlist loaded\n");
	return 0;
    }

    xmp_init (0, NULL, &opt);
    opt.verbose = 0;
    xmp_register_event_callback (player_callback);

    if (xmp_open_audio (&opt) < 0) {
	fprintf (stderr, "Can't open audio device");
	return 0;
    }

    if ((mod_length = xmp_load_module (song)) < 0) {
	fprintf (stderr, "Error loading file: %s\n", song);
	xmp_close_audio();
	return 0;
    }

    xmp_get_module_info (&mi);
    seconds = 0;
    if (strlen (mi.title)) {
	strcpy ((char *)(strchr (cur_item->name, '-') + 2), mi.title);
    }
    cur_item->length = mod_length / 1000;

    pthread_create (&player_thread, NULL, (void *) mod_thread, NULL);
    return 1;
}


void pause_it(void)
{
   xmp_player_ctl(XMP_MOD_PAUSE);
}


void mod_thread (void)
{
    mod_playing = TRUE;
    mod_init = TRUE;
    pthread_setcancelstate (PTHREAD_CANCEL_ENABLE, NULL);
    pthread_setcanceltype (PTHREAD_CANCEL_DEFERRED, NULL);
    xmp_play_module (&mod_length);
    end_song = TRUE;
    mod_playing = FALSE;
    while (1)
       pthread_testcancel();
}


void stop(void)
{
    if (!mod_init)
       return;
    xmp_stop_module();
    while (mod_playing) 
       usleep(50); 
    pthread_cancel(player_thread);
    xmp_close_audio();
    end_song = FALSE;
    mod_init = FALSE;
}


int get_second(void)
{
   return seconds;
}


int get_length(char *song)
{
    return mod_length / 1000;
}

void set_name(struct playlist_item *song)
{  
   strcpy(song->name, (char *)strrchr(song->path, '/') + 1);
}

int is_file(char *path)
{
   char *temp = NULL;
   if (!path)
      return 0;
   if ((temp = strrchr(path,'.'))) {
      if (!strcasecmp((temp+1),"xm")) 
	 return 1;
      if (!strcasecmp((temp+1),"mod"))
	 return 1;
      if (!strcasecmp((temp+1),"s3m"))
	 return 1;
      if (!strcasecmp((temp+1),"stm"))
	 return 1;
      if (!strcasecmp((temp+1),"mtm"))
	 return 1;
      if (!strcasecmp((temp+1),"ptm"))
	 return 1;
      if (!strcasecmp((temp+1),"okt"))
	 return 1;
      if (!strcasecmp((temp+1),"far"))
	 return 1;
      if (!strcasecmp((temp+1),"wow"))
	 return 1;
      if (!strcasecmp((temp+1),"669"))
	 return 1;
      if (!strcasecmp((temp+1),"alm"))
	 return 1;
      if (!strcasecmp((temp+1),"mod"))
	 return 1;
      if (!strcasecmp((temp+1),"Z"))
	 return 1;
      if (!strcasecmp((temp+1),"zip"))
	 return 1;
      if (!strcasecmp((temp+1),"lha"))
	 return 1;
      if (!strcasecmp((temp+1),"lzh"))
	 return 1;
      if (!strcasecmp((temp+1),"bz2"))
	 return 1;
      if (!strcasecmp((temp+1),"mdz"))
	 return 1;
      if (!strcasecmp((temp+1),"gz"))
	 return 1;
   }
   return 0;
}

void version_info(char *id, char *version, char *copyright)
{
   strcpy(id, "xmp");
   strcpy(version, "1.1.4");
   strcpy(copyright, "1996-1998 Claudio Matsuoka and Hipolito Carraro Jr");
}

void init(void)
{
}

void get_info(struct playlist_item *song)
{
}
